# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-07-24 14:34:35
from .pluginsuite import PLUGIN_CONFIG
from .validstructure import load_valid_structure



# -*- coding: utf-8 -*-
# @Author: Joanne Fan
# @Date:   2024-01-26 17:47:47
# @Last Modified by:   Joanne Fan
# @Last Modified time: 2024-02-10 21:01:46

import pytest


@pytest.fixture()
def WATSON_API_KEY():
    WATSON_API_KEY = "MSgOPTS9CvbADe49nEg4wm8_gxeRuf4FGUmlHS9QqAw3"
    return WATSON_API_KEY


@pytest.fixture()
def WATSON_REGION():
    WATSON_REGION = "dallas"
    return WATSON_REGION


@pytest.fixture()
def WATSON_BASE_LANG_MODEL():
    WATSON_BASE_LANG_MODEL = "en-US_NarrowbandModel"
    return WATSON_BASE_LANG_MODEL


@pytest.fixture()
def WATSON_SETTING(WATSON_API_KEY, WATSON_REGION, WATSON_BASE_LANG_MODEL):
    WATSON_SETTING = {
        "engine": "watson",
        "apikey": WATSON_API_KEY,
        "region": WATSON_REGION,
        "base_model": WATSON_BASE_LANG_MODEL
    }
    return WATSON_SETTING


@pytest.fixture()
def WHISPERX_SETTING():
    WHISPERX_SETTING = {
        "engine": "whisperX",
        "language": "English"
    }
    return WHISPERX_SETTING



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-02-24 10:12:04
# @Email:
# @Last Modified by:   Vivian Li 
# @Last Modified time: 2024-03-03 15:12:54
# @Descriptions: declare abstract class for Converter,
#                a converter convert a source to a list of payload
from abc import ABC, abstractmethod
from enum import Enum
from typing import List
from gailbot.profileManager import ProfileObject
from gailbot.payload.payloadObject import PayloadObject
from gailbot.sourceManager.watcher import Watcher


class ConverterType(Enum):
    MixedDirectory = "Mixed Directory"
    SingleMedia = "Single Media File"
    ConvDir = "Conversation Directory"
    TranscribedResult = "Previous GailBot Output"

    def __str__(self):
        return self.value


class Converter(ABC):
    accepted_forms = (
        []
    )  # a list of accepted formats that can be converted to a payload by the converter
    converter_type: ConverterType

    @staticmethod
    @abstractmethod
    def is_accepted_form(path: str) -> bool:
        """
        Return true if path is a file with the accepted form
        """
        pass

    @staticmethod
    @abstractmethod
    def convert(
            path: str, output_dir: str, profile: ProfileObject, watchers: List[Watcher]
    ) -> List[PayloadObject]:
        """
        Given the path to the original source,  the targeted output_dir of the
        transcribed result, the profile applied to the source, and the source
        watchers, convert the source to a list of payload
        """
        pass



# -*- coding: utf-8 -*-
# @Author: Vivian Li 
# @Date:   2024-03-29 17:10:22
# @Last Modified by:   Vivian Li 
# @Last Modified time: 2024-04-02 18:26:36
from typing import Dict, Optional

from gailbot.engineManager.engine.engine import Engine
from gailbot.engineManager.engine.engineProvider import EngineProvider
from gailbot.engineManager.engine.whisper import Whisper
from gailbot.engineManager.engine.whisper.whisperSetting import WhisperSetting


class WhisperProvider(EngineProvider):

    @staticmethod
    def load(name, data: Dict[str, str]) -> Optional[EngineProvider]:
        try:
            engine = WhisperProvider(name, data)
            assert engine.data.engine.lower() == "whisper"
            return engine
        except Exception as e:
            return None

    def __init__(self, name, data: Dict[str, str]):
        self.name = name
        self.data: WhisperSetting = WhisperSetting(**data)

    def make_engine(self) -> Engine:
        return Whisper(self.data)

    def is_cpu_intensive(self) -> bool:
        return True



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-03-29 17:10:22
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-02 18:27:19
from typing import Dict, Optional
import os
from gailbot.engineManager.engine.engineProvider import EngineProvider
from gailbot.engineManager.engine.whisperX.whisperX import WhisperX
from gailbot.engineManager.engine.whisperX.whisperXSetting import WhisperXSetting
from gailbot.shared.utils.logger import makelogger


class WhisperXProvider(EngineProvider):
    @staticmethod
    def load(name, data: Dict[str, str]) -> Optional[EngineProvider]:
        try:
            num_threads = int(data.get("num_threads", 1))
            os.environ["OMP_NUM_THREADS"] = str(num_threads)
            os.environ["MKL_NUM_THREADS"] = str(num_threads)
            os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE" 

            engine = WhisperXProvider(name, data)
            assert engine.data.engine.lower() == "whisperx"
            return engine
        except Exception as _:
            return None

    def __init__(self, name, data: Dict[str, str]):
        self.data: WhisperXSetting = WhisperXSetting(**data)
        self.name = name

    def make_engine(self) -> WhisperX:
        return WhisperX(self.data)

    def is_cpu_intensive(self) -> bool:
        return True



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-02-24 10:14:11
# @Email:
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-02-28 19:06:07
# @Descriptions
import os.path
from .payloadObject import PayloadObject, OutputFolder
from ...shared.utils.general import copy
from ...shared.utils.logger import makelogger

logger = makelogger("transcribed directory loader")


class TranscribedDirPayload(PayloadObject):
    def merge_audio(self, output) -> str:
        for root, dirs, files in os.walk(self.data_files[0]):
            for file in files:
                file_name = os.path.basename(file)
                if file_name == OutputFolder.MERGED_FILE:
                    merged_path = copy(
                        os.path.join(root, file), os.path.join(self.output_dir_struct.media, OutputFolder.MERGED_FILE)
                    )
                    return merged_path
        logger.warning(f"{OutputFolder.MERGED_FILE} is not found")
        return ""

    def stt_transcribe(self):
        print("IN stt transcribe in transcribedDirPayload\n")
        return self.transcription_result.load_result(
            OutputFolder.get_transcribe_dir(self.data_files[0])
        )



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-07-24 14:34:31
import os
from dataclasses import dataclass
from dict_to_dataclass import field_from_dict, DataclassFromDict
import toml


@dataclass
class LogConfig(DataclassFromDict):
    formatter: str = field_from_dict()
    sub_dir_prefix: str = field_from_dict()
    log_dir: str = field_from_dict()


def load_log_config(path):
    d = toml.load(path)
    return LogConfig.from_dict(d["log"])



import os.path
import unittest

from gailbot import FailPluginSuiteRegister
from gailbot.pluginSuiteManager.suite.pluginData import ConfModel
from gailbot.pluginSuiteManager.suite.suite import PluginSuite
from gailbot.shared.utils.general import read_toml
# from test.suiteManager import SUITE_TEST_SRC_PATH, SUITE_TEST_WORKSPACE_PATH
from gailbot.workspace.manager import WorkspaceManager



class PluginSuiteTest(unittest.TestCase):
    root = ""


    def test_create_suite(self):
        workspace = WorkspaceManager()
        workspace.init_workspace()
        conf_model = ConfModel(**read_toml(os.path.join(workspace.suites, "Hello/suite_info.toml")))
        root = workspace.suites
        suite = PluginSuite(conf_model, self.root)
        print("Suite Name: " +  suite.suite_name)
        print(suite.conf_model)
        print(suite.plugins)
        print(suite.dependency_map)
        assert suite.suite_name == "Hello"
        



    # def test_create_legal_suite(self):
    #     conf_model = ConfModel(**read_toml(os.path.join(SUITE_TEST_SRC_PATH, "DummyTestSuite/config.toml")))
    #     suite = PluginSuite(conf_model, self.root)
    #     print(suite.required_plugins)
    #     assert suite.suite_name == "DummyTestSuite"

    # def test_fail_due_to_plugin_not_exist(self):
    #     conf_model = ConfModel(**read_toml(os.path.join(SUITE_TEST_SRC_PATH, "NameNotFound/config.toml")))
    #     try:
    #         PluginSuite(conf_model, self.root)
    #         assert False
    #     except Exception as e:
    #         print(e.__class__)
    #         assert isinstance(e, FailPluginSuiteRegister)

    # def test_fail_due_to_unresolved_dependency(self):
    #     conf_model = ConfModel(**read_toml(os.path.join(SUITE_TEST_SRC_PATH, "UnresolvedDependency/config.toml")))
    #     try:
    #         PluginSuite(conf_model, self.root)
    #         assert False
    #     except Exception as e:
    #         print(e.__class__)
    #         assert isinstance(e, FailPluginSuiteRegister)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-08-01 05:34:44
import os
import re
import boto3
import requests
from botocore.exceptions import ParamValidationError
from cryptography.fernet import Fernet
from typing import List, IO, TextIO
from abc import ABC
from .directoryloader import PluginDirectorySuiteLoader
from gailbot.shared.utils.logger import makelogger
from gailbot.shared.utils.general import read_toml, delete
from gailbot.configs import PLUGIN_CONFIG
from gailbot.shared.utils.download import download_from_urls
from gailbot.pluginSuiteManager.S3BucketManager import S3BucketManager

from ..suite.pluginData import ConfModel
from ..suite.suite import PluginSuite

logger = makelogger("url_loader")


class UrlLoader(ABC):
    """base class for loading plugin from url"""

    def __init__(self, download_dir, suites_dir) -> None:
        self.download_dir = download_dir
        self.suites_dir = suites_dir
        self.dir_loader = PluginDirectorySuiteLoader(suites_dir)
        super().__init__()

    @staticmethod
    def is_supported_url(url: str) -> bool:
        """
        check if the url is supported
        """
        raise NotImplementedError

    def load(self, url: str, f: IO) -> List[PluginSuite]:
        """load the source from the url"""
        raise NotImplementedError


class ZipUrlLoader(UrlLoader):
    """load plugin from an url source"""

    def __init__(self, download_dir, suites_dir) -> None:
        """initialize the plugin loader

        Args:
            download_dir (str): path to where the plugin suite will be downloaded
            suites_dir (str): path to where the plugin will be stored after
                              download
        """
        super().__init__(download_dir, suites_dir)

    @staticmethod
    def is_supported_url(url: str) -> bool:
        try:
            response = requests.head(url)
            content_type = response.headers.get('content-type', '')
            # Check if the content type indicates a zip file
            return content_type.lower() == 'application/zip' or url.lower().endswith('.zip')
        except requests.RequestException as e:
            return False

    def load(self, url: str, f: TextIO) -> [PluginSuite]:
        if not self.is_supported_url(url):
            return []

        download_path = download_from_urls(
            urls=[url], download_dir=self.download_dir, unzip=True
        )[0]

        suite_name = None

        # get the suite name from the toml file
        for root, dirs, files in os.walk(download_path):
            if PLUGIN_CONFIG.CONFIG in files:
                config = os.path.join(root, PLUGIN_CONFIG.CONFIG)
                suite_name = ConfModel(**read_toml(config)).suite.name
                suite_path = download_path[:(download_path.rfind("/"))] + "/" + suite_name
                os.rename(download_path, suite_path)
                logger.info(f"download path: {download_path} suite path: {suite_path}")
                break

        if not suite_name:
            f.write(f"{url} does not store a valid {PLUGIN_CONFIG.CONFIG} configuration file\n")
            return []
        if not suite_path:
            f.write(f"{suite_path} does not exist\n")
            return []

        suites = self.dir_loader.load(suite_path, f)
        delete(suite_path)
        return suites


class S3BucketLoader(UrlLoader):
    """load plugin from an url source"""

    def __init__(self, download_dir, suites_dir) -> None:
        """initialize the plugin loader

        Args:
            download_dir (str): path to where the plugin suite will be downloaded
            suites_dir (str): path to where the plugin will be stored after
                              download
        """
        super().__init__(download_dir, suites_dir)
        self.download_dir = download_dir
        self.suites_dir = suites_dir

    @staticmethod
    def is_supported_url(bucket: str) -> bool:
        try:
            s3BucketManager = S3BucketManager.get_instance()
            r3 = s3BucketManager.get_r3()
            r3.meta.client.head_bucket(Bucket=bucket)
            return True
        except ParamValidationError:
            return False
        except Exception as e:
            return False

    def load(self, bucket: str, f: TextIO) -> List[PluginSuite]:
        if not self.is_supported_url(bucket):
            return []
        

        s3BucketManager = S3BucketManager.get_instance()
        s3 = s3BucketManager.get_s3()

        plugin_suites = []
        # get all object from teh bucket

        objects = s3.list_objects_v2(Bucket=bucket)["Contents"]
        for obj in objects:
            key = obj["Key"]
            if "zip" in key:
                # Generate a pre-signed URL for the object
                url = s3.generate_presigned_url(
                    "get_object", Params={"Bucket": bucket, "Key": key}, ExpiresIn=3600
                )
                url = url[0: url.index("?")]
                logger.info(f"loading plugin from url {url}")
                ZipLoader = ZipUrlLoader(self.download_dir, self.suites_dir)
                plugin_suites.extend(ZipLoader.load(url, f))
        return plugin_suites



# -*- coding: utf-8 -*-
# @Author: Vivian Li 
# @Date:   2024-03-29 17:10:22
# @Last Modified by:   Vivian Li 
# @Last Modified time: 2024-05-03 17:17:10
from typing import Dict, Optional

from gailbot.engineManager.engine.engine import Engine
from gailbot.engineManager.engine.engineProvider import EngineProvider
from gailbot.engineManager.engine.watson import Watson
from gailbot.shared.utils.logger import makelogger
from gailbot.engineManager.engine.watson.watsonSetting import WatsonSetting

logger = makelogger("watson setting")
class WatsonProvider(EngineProvider):

    @staticmethod
    def load(name, data: Dict[str, str]) -> Optional[EngineProvider]:
        try:
            engine = WatsonProvider(name, data)
            assert engine.data.engine.lower() == "watson"
            return engine
        except Exception as e:
            logger.warning(e)
            return None

    def is_cpu_intensive(self) -> bool:
        return False

    def __init__(self, name, data: Dict[str, str]):
        self.data: WatsonSetting = WatsonSetting(**data)
        self.name = name
        Watson.validate_api(apikey=self.data.apikey, region=self.data.region)

    def make_engine(self) -> Engine:
        return Watson(self.data)



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-02-24 10:14:02
# @Email:
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-08 22:41:25
# @Descriptions
import os
import threading
from pydub import AudioSegment
from concurrent.futures import ThreadPoolExecutor, wait
from typing import List
from .payloadObject import PayloadObject, Progress
from ...profileManager import ProfileObject
from ...shared.utils.general import get_name, copy
from ...shared.utils.logger import makelogger
from ...shared.utils.media import AudioHandler
from ...sourceManager.watcher import Watcher
from ...workspace.directory_structure import OutputFolder, TemporaryFolder

logger = makelogger("un-transcribed payload")


class untranscribedDirPayload(PayloadObject):

    def __init__(
        self,
        original_source: str,
        data_files: List[str],
        temp_dir_struct: TemporaryFolder,
        output_path: str,
        profile: ProfileObject,
        watchers: List[Watcher],
    ):
        super().__init__(
            original_source, data_files, temp_dir_struct, output_path, profile, watchers
        )
        self.finished_files = 0
        self.finished_files_lock = threading.Lock()

    def merge_audio(self, output) -> str:
        try:
            if len(self.data_files) == 1:
                merged_path = os.path.join(output, OutputFolder.MERGED_FILE)
                copy(src_path=self.data_files[0], tgt_path=merged_path)
            else:
                audio_handler = AudioHandler()
                merged_path = audio_handler.overlay_audios(
                    self.data_files, output, OutputFolder.MERGED_FILE
                )
            return merged_path
        except Exception as e:
            logger.error(e, exc_info=True)

    def stt_transcribe(self) -> bool:
        print("IN stt transcribe in untranscribedDirPayload\n")
        if len(self.data_files) == 1:
            return self.transcribe_file(self.data_files[0])
        else:
            return self.transcribe_files(self.data_files)

    def transcribe_file(self, path: str):
        try:
            for w in self.watchers:
                w.watch(
                    Progress.Transcribing.format(
                        name=self.name, engine=self.engine.get_engine_name()
                    )
                )
            result = self.engine.transcribe(path, self.temp_dir_struct.transcribe_ws)
            assert self.transcription_result.save_data({get_name(path): result})
            return True
        except Exception as e:
            logger.error(e, exc_info=True)
            return False

    def transcribe_files(self, paths: List[str]):
        total_time = 0
        time_loaded = True
        for p in paths:
            try:
                audio: AudioSegment = AudioSegment.from_file(p)
                total_time += audio.duration_seconds
            except Exception as e:
                logger.warn(f"Failed to get the audio time", exc_info=True)
                time_loaded = False
        logger.info(f"The total audio time is {total_time} seconds")

        def report_finish(_):
            with self.finished_files_lock:
                self.finished_files += 1
                for w in self.watchers:
                    w.watch(
                        f"{self.name}: Transcribed {self.finished_files} / {len(self.data_files)} files."
                    )

        try:
            for watcher in self.watchers:
                watcher.watch(
                    Progress.Transcribing.format(
                        name=self.name, engine=self.engine.get_engine_name()
                    )
                )
            futures = dict()
            results = dict()
            with ThreadPoolExecutor(min(3, len(paths))) as executor:
                for file in paths:
                    future = executor.submit(self.transcribe_wrapper, file)
                    future.add_done_callback(report_finish)
                    futures[get_name(file)] = future
                if time_loaded:
                    wait(futures.values(), timeout=total_time * 2)
                else:
                    wait(futures.values())

            for name, future in futures.items():
                result = future.result()
                assert result != False
                results[name] = result
            assert self.transcription_result.save_data(results)
            return True
        except Exception as e:
            logger.error(e, exc_info=True)
            return False

    def transcribe_wrapper(self, file):
        try:
            return self.engine.transcribe(file, self.temp_dir_struct.transcribe_ws)
        except Exception as e:
            logger.error(e, exc_info=True)
            return False



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Joanne Fan
# @Last Modified time: 2024-04-01 02:20:18
# @Description: Any file accepted by gailbot will be initially stored as an instance of SourceObject
#               SourceObject will store the Profile and Loader applied to the Source,
#               A set of SourceObjects will be managed by SourceManager
from typing import List

from gailbot.payload.payloadConverter.mediaConverter import MediaConverter
from gailbot.payload.payloadConverter.conversationDirConverter import ConversationDirConverter
from gailbot.payload.payloadConverter.mixedDirConverter import MixedDirectoryConverter
from gailbot.payload.payloadConverter.transcribedDirConverter import TranscribedDirConverter
from gailbot.payload.payloadObject import PayloadObject
from gailbot.profileManager import ProfileObject
from gailbot.sourceManager.watcher import Watcher


class SourceObject:
    converters = [TranscribedDirConverter,
                  MediaConverter,
                  ConversationDirConverter,
                  MixedDirectoryConverter]

    def __init__(self, source_id, source_path, output):
        self.source_id = source_id
        self.name = source_id
        self.source_path = source_path
        self.output_dir = output
        self.profile_name: str | None = None
        self.profile_obj: ProfileObject | None = None
        self.watchers: List[Watcher] = []
        self.converter = None
        for converter in self.converters:
            if converter.is_accepted_form(source_path):
                self.converter = converter
                break

    def link_profile(self, profile_name: str):
        """
        link the source to a profile name

        Parameters
        ----------
        profile_name : str

        Returns
        -------

        """
        self.profile_name = profile_name

    def initialize_profile(self, profile: ProfileObject):
        """
        initialize the profile object used by the source, this function must be called
        before converting the source to a payload for transcription

        Parameters
        ----------
        profile

        Returns
        -------

        """
        self.profile_obj = profile

    def add_watcher(self, watcher: Watcher):
        self.watchers.append(watcher)

    def compatible_converters(self):
        compatible = [c for c in self.converters if c.is_accepted_form(self.source_path)]
        return compatible

    def change_converter(self, converter):
        if converter.is_accepted_form(self.source_path):
            self.converter = converter
            return True
        else:
            return False

    def convert(self) -> List[PayloadObject]:
        if not self.converter:
            return []
        return self.converter.convert(self.source_path, self.output_dir, self.profile_obj, self.watchers)



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-04-09 17:02:48
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-24 19:08:53
from typing import List
from pydantic import BaseModel

from gailbot.configs import whisper_config_loader
from gailbot.setting_interface.formItem import FormItem, FormType


class WhisperSetting(BaseModel):
    engine: str
    language: str
    detect_speakers: bool = False

    @staticmethod
    def predefined_config():
        config = whisper_config_loader()
        return config

    @staticmethod
    def get_setting_config() -> List[FormItem]:
        return [
            FormItem(type=FormType.OnOff, name="detect_speakers", default_value= False),
            FormItem(type=FormType.Selection, name="language", selection_items=["af", "am", "ar", "as", "az", "ba", "be", "bg", "bn", "bo", "br", "bs", "ca", "cs", 
            "cy", "da", "de", "el", "en", "es", "et", "eu", "fa", "fi", "fo", "fr", "gl", "gu", 
            "ha", "haw", "he", "hi", "hr", "hu", "hy", "id", "is", "it", "ja", "jw", "ka", 
            "kk", "km", "kn", "ko", "la", "lb", "ln", "lo", "lt", "lv", "mg", "mi", "mk", 
            "ml", "mn", "mr", "ms", "mt", "my", "ne", "nl", "no", "nn", "oc", "pa", "pl", 
            "ps", "pt", "ro", "ru", "sa", "sd", "si", "sk", "sl", "sn", "so", "sq", "sr", 
            "sv", "sw", "ta", "te", "tg", "th", "tl", "tr", "tt", "uk", "ur", "uz", "vi", 
            "yi", "yo", "yue", "zh"], default_value= "en"),
        ]  # TODO: get the list of selectable language



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Lakshita Jain
# @Last Modified time: 2023-11-05 09:30:39
import os
from datetime import date
from pathlib import Path
import logging
import userpaths
from gailbot.configs import log_config_loader

LOG_CONFIG = log_config_loader()


def makelogger(logger_name: str) -> logging.Logger:
    """create a logger that can be identified by logger_name, and the logger's
        message will be stored into its own individual file inside the backend
        folder

    Args:
        logger_name (str): the name of lof files

    Returns:
        logger: the logger
    """
    # get the all existing log handler
    root_logger = logging.getLogger()
    handlers = root_logger.handlers

    # get the named_log handler
    logger = logging.getLogger(logger_name)
    for handler in handlers:
        logger.addHandler(handler)

    logger.setLevel(logging.DEBUG)
    # each backend log files one each date will be stored in a separate folder
    today = date.today()
    date_log_dir = os.path.join(
        userpaths.get_profile(),
        LOG_CONFIG.log_dir,
        f"{LOG_CONFIG.sub_dir_prefix}-{today}",
    )
    if not Path(date_log_dir).is_dir():
        os.makedirs(date_log_dir)
    path = os.path.join(date_log_dir, f"GailBot-{logger_name}-{today}.log")
    formatter = logging.Formatter(LOG_CONFIG.formatter)
    file_handler = logging.FileHandler(path)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    return logger



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-06 15:49:53
# @Last Modified by:   Muhammad Umair
# @Last Modified time: 2023-01-09 11:21:54

from abc import ABC, abstractmethod
from typing import List, TypedDict


class UttDict(TypedDict):
    """
    Defines a class for the utterance dictionary
    """

    speaker: str
    start: float
    end: float
    text: str


class Engine(ABC):
    """
    A wrapper for third party STT engine API, provides a consistent set of methods
    to perform engine initialization and transcription
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def __repr__(self):
        """
        Returns all the configurations and additional metadata
        """
        raise NotImplementedError()

    @abstractmethod
    def transcribe(self, audio_path, payload_workspace) -> List[UttDict]:
        """
        Use the engine to transcribe an item
        """
        pass

    @abstractmethod
    def was_transcription_successful(self) -> bool:
        """ 
        Return true if the transcription is successful 
        """
        pass

    @abstractmethod
    def get_engine_name(self) -> str:
        """
        Obtain the name of the current engine.

        Returns:
            (str): Name of the engine.
        """
        pass

    @abstractmethod
    def get_supported_formats(self) -> List[str]:
        """
        Obtain a list of audio file formats that are supported.

        Returns:
            (List[str]): Supported audio file formats.
        """
        pass

    @abstractmethod
    def is_file_supported(self, file_path: str) -> bool:
        """
        Determine if the given file is supported by the engine.

        Args:
            file_path (str)

        Returns:
            (bool): True if file is supported. False otherwise.
        """
        pass



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-26 16:16:54
# @Last Modified by:   Joanne Fan
# @Last Modified time: 2024-02-10 21:01:46
import pytest
import os

from gailbot.engineManager.engineManager import EngineManager
from gailbot.shared.exception.serviceException import EngineNotFound, DuplicateProfile, RemoveDefaultEngine
from gailbot.shared.utils.general import is_file


@pytest.fixture()
def engineWorkspace(tmp_path):
    temp_ws = tmp_path / "temp_workspace"
    temp_ws.mkdir()
    return temp_ws


@pytest.fixture()
def eManager(engineWorkspace):
    return EngineManager(engineWorkspace)


@pytest.fixture()
def eManagerWhisperx(eManager, WHISPERX_SETTING):
    eManager.add_engine("test_whisperX_engine", WHISPERX_SETTING)
    return eManager


def test_remove_default_engine(eManager):
    with pytest.raises(RemoveDefaultEngine):
        eManager.remove_engine(eManager.default_name)


def test_add_engine(eManager, WATSON_SETTING, WHISPERX_SETTING):
    assert eManager.add_engine("test_watson_engine", WATSON_SETTING)
    assert eManager.is_engine("test_watson_engine")

    assert eManager.add_engine("test_whisperX_engine", WHISPERX_SETTING)
    assert eManager.is_engine("test_whisperX_engine")

    eManagerB = EngineManager(eManager.workspace)
    assert eManagerB.is_engine("test_watson_engine")
    assert eManagerB.is_engine("test_whisperX_engine")


def test_add_engine_fail(eManagerWhisperx, WHISPERX_SETTING):
    with pytest.raises(DuplicateProfile):
        eManagerWhisperx.add_engine("test_whisperX_engine", WHISPERX_SETTING)


def test_remove_engine(eManagerWhisperx):
    engine_name = "test_whisperX_engine"
    eManagerWhisperx.remove_engine(engine_name)

    assert not eManagerWhisperx.is_engine(engine_name)
    assert engine_name not in eManagerWhisperx.available_engines

    path = os.path.join(eManagerWhisperx.workspace, f"{engine_name}.toml")
    assert not is_file(path)


def test_remove_engine_fail(eManager):
    with pytest.raises(EngineNotFound):
        eManager.remove_engine("test_whisperX_engine")


def test_get_engine_provider(eManager):
    provider = eManager.get_engine_provider(eManager.default_name)
    assert provider.engine_data() == eManager.default_data


def test_get_engine_provider_fail(eManager):
    with pytest.raises(EngineNotFound):
        engine_provider = eManager.get_engine_provider("test_whisperX_engine")


def test_get_engine_data(eManagerWhisperx, WHISPERX_SETTING):
    # Default engine data
    engine_data = eManagerWhisperx.get_engine_setting(eManagerWhisperx.default_name)
    assert engine_data == eManagerWhisperx.default_data

    # WhisperX engine data
    engine_data = eManagerWhisperx.get_engine_setting("test_whisperX_engine")
    assert engine_data == WHISPERX_SETTING


def test_get_engine_data_fail(eManager):
    with pytest.raises(EngineNotFound):
        engine_data = eManager.get_engine_setting("test_whisperX_engine")


def test_update_engine(eManagerWhisperx, WHISPERX_SETTING):
    new_setting = WHISPERX_SETTING.copy()
    new_setting["language"] = "french"
    assert eManagerWhisperx.update_engine("test_whisperX_engine", new_setting)
    assert eManagerWhisperx.get_engine_setting("test_whisperX_engine") == new_setting


def test_update_engine_fail(eManager):
    with pytest.raises(EngineNotFound):
        eManager.update_engine("test_whisperX_engine", {})



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-02-15 22:13:22
# @Description: abstract class that defines a list of method must be implemented
#  by a Plugin
from abc import ABC
from typing import Any, Dict
from gailbot.shared.utils.logger import makelogger
from gailbot.pluginSuiteManager.suite.gbPluginMethod import Methods


class Plugin(ABC):
    """
    Abstract class that defines an interface for a plugin
    """

    def __init__(self) -> None:
        self.logger = makelogger(f"Plugin-{self.__class__}")
        self.name = self.__class__
        self.successful = False

    @property
    def is_successful(self) -> bool:
        return self.successful

    def apply(
        self, dependency_outputs: Dict[str, Any], methods: Methods, *args, **kwargs
    ) -> Any:
        """
        Wrapper for plugin algorithm that has access to dependencies =,
        Args:
            dependency_outputs (Dict[str,Any]):
                Mapping from all plugins this plugin is depended on and their
                outputs.
            methods (Methods):
                Methods object that stores data for transcription
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"plugin {self.name}"



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-04-24 18:59:42
# @Last Modified by:   Your name
# @Last Modified time: 2024-04-24 20:57:01
from enum import Enum
from typing import Any
from dataclasses import dataclass

class FormType(Enum):
    Text = 1
    Selection = 2
    OnOff = 3
    File = 4
    MultiSelect = 5
    Number = 6

#
@dataclass(init=True)
class FormItem:
    type: FormType
    name: str
    selection_items: Any | None = None
    default_value: Any | None = None
    min: int | None = None          
    max: int | None = None      



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-06 23:58:52
# @Description: A plugin suite contains multiple plugins. A PluginSuite
# object stores every information about a suite, including the dependencies between
# each plugins, suite metadata , suite documentation path, suite format markdown path.
# When itself is called, it execute procedure to run the suite.
import sys
import os
from collections import defaultdict
from typing import Dict, List, Any, Tuple, Optional
import subprocess
import os.path
import yaml   
import shutil
import xml.etree.ElementTree as ET

from gailbot.shared.exception.serviceException import FailPluginSuiteRegister
from gailbot.pluginSuiteManager.error.errorMessage import SUITE_REGISTER_MSG
from gailbot.pluginSuiteManager.suite.gbPluginMethod import GBPluginMethods
from gailbot.configs import PLUGIN_CONFIG
from gailbot.pluginSuiteManager.suite.pluginData import Suite, ConfModel, Requirements, Dependencies
from gailbot.shared.pipeline import (
    Pipeline,
)
from gailbot.shared.utils.general import get_name
from pathlib import Path
import importlib
import platform
from gailbot.pluginSuiteManager.suite.pluginComponent import PluginComponent
from gailbot.shared.utils.logger import makelogger
from gailbot.shared.utils.general import read_toml
from gailbot.workspace.manager import WorkspaceManager
from gailbot.workspace.directory_structure import OutputFolder

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from gailbot.pluginSuiteManager.APIConsumer import APIConsumer
from S3BucketManager import S3BucketManager
from userpaths import get_profile
from gailbot.shared.utils.download import download_single_url
import platform
import threading, signal

USER = get_profile()

logger = makelogger("plugin suite")

import uuid

def get_real_username():
    try:
        return subprocess.check_output(["stat", "-f%Su", "/dev/console"], text=True).strip()
    except Exception:
        return os.environ.get("USER") or os.environ.get("LOGNAME") or os.getlogin()

def get_gailbot_root_path() -> str:
    username = get_real_username()
    system = platform.system()

    if system == "Darwin":  # macOS
        base_path = os.path.join("/Users", username)
    elif system == "Linux":
        base_path = os.path.join("/home", username)
    else:
        raise EnvironmentError(f"Unsupported OS: {system}")

    if os.path.isdir(base_path):
        return base_path

    # Fall back to full scan if not found
    for root, dirs, files in os.walk("/", topdown=True):
        dirs[:] = [d for d in dirs if not d.startswith(".") and "Volumes" not in root and "private" not in root]
        if os.path.basename(root) == username:
            if os.path.isdir(root):
                return root

    raise FileNotFoundError(f"GailBot directory not found under expected paths for {username}")

class PluginSuite:
    """
    Manages a suite of plugins and responsible for loading, queries, and
    execution.
    Needs to store the details of each plugin (source file etc.)
    """

    def __init__(self, conf_model: ConfModel, root: str):
        """a dictionary of the dependency map  -> transcriptionPipeline argument"""
        self.suite_name = conf_model.suite.name
        self.conf_model = conf_model
        self.source_path = root
        self.optional_plugins: List[str] = []
        self.required_plugins: List[str] = []
        self.workspace = WorkspaceManager()
        self.workspace.init_workspace()
        # suite and document_path will be loaded in _load_from_config
        self.suite = conf_model.suite
        self.formatmd_path = os.path.join(root, self.suite_name, PLUGIN_CONFIG.FORMAT)
        self.dependency_map, self.plugins = self._load_from_config(conf_model, root)

        # self.download_plugins(plugin_ids= self.plugins)

        # Add vars here from conf.
        self._is_ready = True
        self.is_official = False

    @property
    def name(self) -> str:
        return self.suite_name

    @property
    def is_ready(self):
        return self._is_ready

    def set_to_official_suite(self):
        """set the plugin to official plugin"""
        self.is_official = True

    def __repr__(self):
        return (
            f"Plugin Suite: {self.name}\n" f"Dependency map: {self.dependency_graph()}"
        )

    def mountable_or_root(self, candidate: str, root: str) -> str:
 
        test_cmd = [
            "docker", "run", "--rm",
            "-v", f"{candidate}:/mnt:rw",
            "alpine", "true"                 # tiny base image always present
        ]
        ok = subprocess.run(
            test_cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ).returncode == 0

        if ok:
            return candidate

        fallback = Path(root, "gailbot_work")
        fallback.mkdir(parents=True, exist_ok=True)
        return str(fallback)

    def __call__(self, base_input: Any, methods: GBPluginMethods, selected_plugins=None) -> Dict:
        selected_plugins = self.plugins
        output_path = OutputFolder(methods.out_path)

        adj_list = {k: v for k, v in self.dependency_map.items() if k in selected_plugins}
        dag = self._topological_sort(adj_list)

        host_id = next(pid for pid, deps in self.dependency_map.items() if not deps)
        host_base_path = selected_plugins[host_id]

        host_path = os.path.join(self.workspace.plugins, host_base_path)
        os.makedirs(os.path.join(host_path, "transcript"), exist_ok=True)
        with open(os.path.join(host_path, "transcript", "dag.txt"), "w") as f:
            f.write(" ".join(dag))
        xml = self._convert_to_xml(methods.utterances)
        with open(os.path.join(host_path, "transcript", "original_data.txt"), "w") as f:
            f.write(ET.tostring(xml, encoding="unicode", method="xml"))
        os.makedirs(output_path.transcribe, exist_ok=True)
        with open(os.path.join(output_path.transcribe, "original_trancript.xml"), "w") as f:
            f.write(ET.tostring(xml, encoding="unicode", method="xml"))

        root_path = get_gailbot_root_path()
        project_name = uuid.uuid4().hex[:128]

        methods.work_path = self.mountable_or_root(methods.work_path, root_path)
        os.makedirs(methods.work_path, exist_ok=True)
        with open(os.path.join(methods.work_path, ".write_test"), "w") as f:
            f.write("ok")

        env = os.environ.copy()
        env["WORK_PATH"] = methods.work_path
        env["OUTPUT"]    = str(output_path.root)
        env["NAME"]      = project_name
        env["ROOT"]      = root_path

        system = platform.system()
        if system == "Linux":
            pagesize   = os.sysconf("SC_PAGESIZE")
            phys_pages = os.sysconf("SC_PHYS_PAGES")
            total_bytes = pagesize * phys_pages
        elif system == "Darwin":
            total_bytes = int(subprocess.check_output(["sysctl","-n","hw.memsize"]))
        else:
            total_bytes = 1 << 30
        shm_mb  = total_bytes // (2 * 1024 * 1024)
        threads = os.cpu_count() or 1

        compose_file = os.path.join(self.source_path, self.suite_name, "plugin_suite_composed.yaml")
        cfg = yaml.safe_load(open(compose_file))
        for svc in cfg["services"].values():
            svc["shm_size"] = f"{shm_mb}m"
            svc.setdefault("environment", {})["OMP_NUM_THREADS"] = str(threads)
        yaml.safe_dump(cfg, open(compose_file, "w"), sort_keys=False)

        subprocess.run(
            ["docker","compose","-f",compose_file,"-p",project_name,
            "up","--build","-d"],
            env=env, check=True
        )
        log_proc = subprocess.Popen(
            ["docker", "compose", "-f", compose_file, "-p", project_name,
            "logs", "--no-color", "--timestamps", "-f"],
            env=env,
            stdout=subprocess.PIPE,      # capture stdout
            stderr=subprocess.STDOUT,    # merge stderr in
            text=True,
        )

        def _forward_logs():
            for line in log_proc.stdout:
                logger.info(line.rstrip())

        t = threading.Thread(target=_forward_logs, daemon=True)
        t.start()

        service_names = list(cfg["services"].keys())
        subprocess.run(
            ["docker","compose","-f",compose_file,"-p",project_name,
            "wait"] + service_names,
            env=env, check=True,
            text=True
        )
        log_proc.send_signal(signal.SIGINT)
        log_proc.wait()

        subprocess.run(
            ["docker","compose","-f",compose_file,"-p",project_name,
            "down","--rmi","all","--volumes","--remove-orphans"],
            env=env, check=True,
            
        )


        os.makedirs(output_path.analysis, exist_ok=True)

        # 1) Collect all plugin result files into results dict
        results: Dict[str, Dict[str, str]] = defaultdict(dict)
        for item in os.listdir(methods.work_path):
            src_path = os.path.join(methods.work_path, item)
            if os.path.isfile(src_path) and item.startswith("plugin_") and "_result." in item:
                # parse plugin_id and extension
                base_name = os.path.basename(item)               # plugin_{id}_result.ext
                parts = base_name[:43].split("_")
                plugin_id = parts[1]
                ext = os.path.splitext(item)[1].lstrip(".")
                # read content
                with open(src_path, "r", encoding="utf-8") as f:
                    results[plugin_id][ext] = f.read()

        # 2) Write each plugin’s results into its own folder
        for plugin_id, files in results.items():
            plugin_dir = os.path.join(output_path.analysis, f"plugin_{plugin_id}")
            os.makedirs(plugin_dir, exist_ok=True)
            for ext, content in files.items():
                fname = f"plugin_{plugin_id}_result.{ext}"
                dst_path = os.path.join(plugin_dir, fname)
                with open(dst_path, "w", encoding="utf-8") as dst:
                    dst.write(content)
                logger.info(f"Wrote {dst_path}")

        # 3) Also collect the final plugin’s files into a separate folder
        final_plugin_id = dag[-1]
        final_dir = os.path.join(output_path.analysis, "final_plugin")
        os.makedirs(final_dir, exist_ok=True)
        for ext, content in results.get(final_plugin_id, {}).items():
            fname = f"plugin_{final_plugin_id}_result.{ext}"
            dst_path = os.path.join(final_dir, fname)
            with open(dst_path, "w", encoding="utf-8") as dst:
                dst.write(content)
            logger.info(f"Wrote final plugin file {dst_path}")



    def check_required_files(self, directory):
        required_files = {"app.py", "client.py", "utils.py"}
        existing_files = set(os.listdir(directory))

        return required_files.issubset(existing_files)

    def is_plugin(self, plugin_name: str) -> bool:
        """given a name , return true if the plugin is in the plugin suite"""
        return plugin_name in self.plugins

    def plugin_names(self) -> List[str]:
        """Get names of all plugins"""
        return list(self.plugins.keys())

    def dependency_graph(self) -> Dict:
        """Return the entire dependency graph as a dictionary"""
        return self.dependency_map

    def get_meta_data(self) -> Suite:
        """get the metadata about this plugin"""
        return self.suite



    ##########
    # PRIVATE
    ##########
    def _load_from_config(
        self, conf_model: ConfModel, root: str
    ) -> Tuple[Dict[str, List[str]], Dict[str, str]] | None:
        """
        load the plugin suite, the information about each plugin name,
        and its path is stored in the dict_config, all path information
        is relative to the abs_path

        Parameters
        ----------
        conf_model: stores the plugin suite data for suite registration
        root: the path to the root folder of the plugin suite source code

        """
        plugins: Dict[str, str] = dict()

        plugin_entries = conf_model.suite.plugins.split(' ')

        for plugin_id in plugin_entries:
            plugin_path = os.path.join(self.workspace.plugins, plugin_id)
            plugins[plugin_id] = plugin_path
            
        self.download_plugins(plugin_ids = plugin_entries)
        
        dependency_map: Dict[str, List[str]] = self._create_adjacency_list(plugins)


        return (
            dependency_map,
            plugins,
        )  
    

    def _create_adjacency_list(self, plugins_data: Dict[str, str]) -> Dict[str, List[str]]:
        """
        Create an adjacency list from the list of plugin data.

        Args:
        - plugins_data: List of dictionaries containing plugin information.

        Returns:
        - Dict: An adjacency list where keys are plugin names and values are lists of dependencies.
        """
        adjacency_list = {}
        for plugin_id, plugin_path in plugins_data.items():
            toml_file_path = os.path.join(plugin_path, "plugin_info.toml")
            

            plugin_info = read_toml(toml_file_path)

            adjacency_list[plugin_id] = list(plugin_info.get('requirements', {}).values())

                    

        return adjacency_list
    

    def _convert_to_xml(self, data):
        root = ET.Element("transcript")
        
        current_speaker = None
        current_u = None
        
        for data in data.values():
            for item in data:
                #logger.info(item)
                if item['speaker'] != current_speaker:
                    current_speaker = item['speaker']
                    current_u = ET.SubElement(root, "u", speaker=current_speaker)
                
                word = ET.SubElement(current_u, "w", start=str(item['start']), end=str(item['end']))
                word.text = item['text']

        return root



    
    def _topological_sort(self, adj_list):
        # Initialize the graph and in-degree dictionary
        graph = {}
        in_degree = {}

        # Build the graph and compute in-degrees of each node
        for node in adj_list:
            if node not in graph:
                graph[node] = []
            if node not in in_degree:
                in_degree[node] = 0
            for dep in adj_list[node]:
                if dep not in graph:
                    graph[dep] = []
                graph[dep].append(node)
                if node not in in_degree:
                    in_degree[node] = 0
                in_degree[node] += 1

        # Find all nodes with in-degree 0
        queue = []
        for node in in_degree:
            if in_degree[node] == 0:
                queue.append(node)
        
        order = []

        # Process nodes with in-degree 0 and update the in-degrees of their neighbors
        while queue:
            current = queue.pop(0)
            order.append(current)
            for neighbor in graph[current]:
                in_degree[neighbor] -= 1
                if in_degree[neighbor] == 0:
                    queue.append(neighbor)

        # Check if there was a cycle
        if len(order) == len(in_degree):
            return order
        else:
            raise ValueError("A cycle was detected in the dependencies")



    
    def _separate_plugins(self, adj_list):
        # Create dictionaries to store the counts of incoming and outgoing edges
        outgoing_edges = defaultdict(set)
        incoming_edges = defaultdict(set)
        
        # Populate the dictionaries with the adjacency list information
        for plugin, dependencies in adj_list.items():
            outgoing_edges[plugin].update(dependencies)
            for dep in dependencies:
                incoming_edges[dep].add(plugin)
        
        # Find independent and required plugins
        self.optional_plugins = [plugin for plugin in outgoing_edges if plugin not in incoming_edges]
        self.required_plugins = [plugin for plugin in incoming_edges]
        


    def sub_dependency_graph(
        self, selected: List[str]
    ) -> Optional[Dict[str, List[str]]]:
        """
        given a selected list of plugins, return a subgraph of the dependency graph that
        include only the required plugin and the list of selected plugin

        Parameters
        ----------
        selected

        Returns
        -------

        """
        selected.extend(self.required_plugins)
        selected = set(selected)
        new_dependency = dict()
        for key, dependency in self.dependency_map.items():
            if key in selected:
                new_dependency[key] = list(
                    filter(lambda elt: elt in selected, dependency)
                )
        if not self.__check_dependency(new_dependency):
            logger.error(f"cannot resolve dependency for graph {new_dependency}")
        return new_dependency

    def __check_dependency(self, graph: Dict[Any, List[Any]]):
        """

        Parameters
        ----------
        graph

        Returns None
        -------

        Raises
        -------
        FailPluginSuiteRegister

        """
        visited = {k: 0 for k in graph.keys()}

        def check_circle(node: Any):
            visited[node] = -1
            for dependency in graph[node]:
                if visited[dependency] == -1:
                    raise FailPluginSuiteRegister(
                        self.suite_name,
                        SUITE_REGISTER_MSG.FAIL_LOAD_PLUGIN.format(
                            plugin=node,
                            cause=f" cannot resolve dependency {dependency} for plugin {node}",
                        ),
                    )
                elif visited[dependency] == 0:
                    if check_circle(dependency):
                        raise FailPluginSuiteRegister(
                            self.suite_name,
                            SUITE_REGISTER_MSG.FAIL_LOAD_PLUGIN.format(
                                plugin=node,
                                cause=f" cannot resolve dependency {dependency} for plugin {node}",
                            ),
                        )
            visited[node] = 1

        for node in graph.keys():
            check_circle(node)

        return True
    
    def download_plugins(self, plugin_ids: List[str]):
        s3 = S3BucketManager.get_instance()
        api = APIConsumer.get_instance()
        for plugin in plugin_ids:
            plugin_info = api.fetch_plugin_info(plugin_id= plugin)
            
            s3_url = plugin_info['s3_url']
            local_dir = os.path.join(self.workspace.plugins, plugin)
            if not os.path.isdir(local_dir):
                os.makedirs(local_dir)
            download_single_url(url= s3_url, download_path= local_dir, extract_root= local_dir, s3_client=s3.get_s3(), unzip= True, true_extract=True)


# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-08-01 05:37:45
from typing import TypedDict, Dict, List

from .resultInterface import ResultInterface
from gailbot.shared.utils.logger import makelogger

logger = makelogger("analysis-result")


class AnalysisResultDict(TypedDict):
    plugin_suite: str
    success: List[str]
    failure: List[str]


class AnalysisResult(ResultInterface):
    """
    Defines a class for the analysis result
    """

    def __init__(self, workspace: str):
        super().__init__(workspace)
        self.data: Dict[str, AnalysisResultDict] = dict()

    def save_data(self, data) -> bool:
        """
        Saves the result data

        Args:
            data: [AnalysisResultDict]: data to save, in the form
            of a dictionary mapping strings to analysis results

        Returns:
            Bool: True if successfully saved, false if not
        """
        self.data = data
        return True

    def output(self, path) -> bool:
        """NOTE: currently no data for analysis will be written"""
        try:
            return True
        except Exception as e:
            logger.error(e, exc_info=e)
            return True

    def get_data(self):
        """
        Accesses object's data

        Returns:
            Data in the form Dict[str, AnalysisResultDict]
        """
        return self.data

    def get_serialized_data(self):
        return {str(k): str(v) for k, v in self.data.items()}



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-09 11:25:49
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-11 00:21:21

import os
from typing import List, Any, Dict, Tuple
from itertools import chain
from copy import deepcopy
from ibm_watson import SpeechToTextV1
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_watson.websocket import AudioSource
from .recognize_callback import CustomWatsonCallbacks
from .recognition_results import RecognitionResult

# local import
from gailbot.shared.utils.logger import makelogger
from gailbot.shared.exception.transcribeException import (
    APIKeyError,
    InternetConnectionError,
    AudioFileError,
    GetUttResultError,
)
from gailbot.configs import watson_config_loader
from gailbot.shared.utils.media import MediaHandler
from gailbot.shared.utils.general import (
    get_extension,
    get_name,
    is_file,
    get_size,
)
from gailbot.engineManager.engine.watson.watsonSetting import WatsonSetting

WATSON_CONFIG = watson_config_loader()
logger = makelogger("watson_core")


class WatsonCore:
    """
    Implement shared functionalities to transcribe an audio file through
    watson STT engine
    """

    def __init__(self, apikey: str, region: str, watson_setting: WatsonSetting):
        self.apikey = apikey
        self.region = region
        self.setting = watson_setting
        self.media_h = MediaHandler()

        # Parse confs
        self._format_to_content_types = WATSON_CONFIG.format_to_content
        self._defaults = WATSON_CONFIG.defaults

        if not WatsonCore.valid_region_api(apikey, region):
            raise APIKeyError()

    @staticmethod
    def valid_region_api(apikey, region):
        """check if an api key is a valid api key under the region

        Args:
            apikey (str): a string that stores the api key
            region (str): the region

        Returns:
            bool: true if the api key is valid, false otherwise
        """
        if region not in WATSON_CONFIG.regions_uris:
            logger.error("region invalid")
            return False
        if not WatsonCore.is_api_valid(apikey, WATSON_CONFIG.regions_uris[region]):
            logger.error("API invalid")
            return False
        return True

    @property
    def supported_formats(self) -> List[str]:
        """
        Access the supported formats of an instance of object WatsonCore

        Returns:
            List[str] : list of supported formats
        """
        return ["wav"]

    @property
    def regions(self) -> Dict:
        """
        Access the regions of an instance of object WatsonCore

        Returns:
            Dict representing the regions
        """
        return WATSON_CONFIG.regions_uris

    @property
    def defaults(self) -> Dict:
        """
        Access the defaults of an instance of object WatsonCore

        Returns:
            Dict representing the defaults
        """
        return self._defaults

    def is_file_supported(self, file: str) -> bool:
        """
        Determines if a given file is supported

        Args:
            file (str) : file name to check if supported

        Returns: True if the file is supported, false if not.
        """
        return get_extension(file) in self.supported_formats

    def transcribe(
        self,
        audio_path: str,
        payload_workspace: str,
        base_model: str,
        language_customization_id: str,
        acoustic_customization_id: str,
        setting: List
    ):
        """
        Transcribes the provided audio stream using a websocket connections.
        And output the result to given output directory

        Args:

        Parameters
        ----------
        audio_path (str) : path to audio file that will be transcribed
        base_language_model (str) :  specifies the base_language_model
        language_customization_id (str) : ID of the custom acoustic model
        acoustic_customization_id (str) : ID of the custom language model.
        """
        # compress the audio if it is too big
        path_copy = audio_path
        try:
            new_path = os.path.join(payload_workspace, os.path.basename(audio_path))
            audio_path = MediaHandler.convert_to_16bit_wav(audio_path, new_path)
        except Exception as e:
            logger.error(e, exc_info=e)
            audio_path = path_copy
        try:
            if get_size(audio_path) >= WATSON_CONFIG.max_file_size_bytes:
                audio_path = MediaHandler.compress_to_opus(
                    audio_path, payload_workspace
                )
                logger.info(
                    f"File size exceeds size limit: compressed audio_path is {audio_path}"
                )
        except Exception as e:
            logger.error("error in compressing the media file", exc_info=e)
            raise AudioFileError()

        recognize_callbacks = CustomWatsonCallbacks()
        try:
            self._websockets_recognize(
                setting,
                audio_path,
                base_model,
                recognize_callbacks,
                language_customization_id,
                acoustic_customization_id,
            )
        except Exception as e:
            logger.error(e, exc_info=e)
            raise InternetConnectionError()

        try:
            audio_name = get_name(audio_path)
            utterances = self._prepare_utterance(
                audio_name, recognize_callbacks.get_results()
            )
            return utterances
        except Exception as e:
            logger.error(e, exc_info=e)
            raise GetUttResultError()

    ###############
    # PRIVATE
    ##############
    def _websockets_recognize(
        self,
        setting: List,
        audio_path: str,
        base_model: str,
        recognize_callbacks: CustomWatsonCallbacks,
        language_customization_id: str = None,
        acoustic_customization_id: str = None,
    ) -> Any:
        """
        Transcribes the provided audio stream using a websocket connections.
        All attributes MUST be set before using this method.

        Args:
            audio_path (str) :
                path to audio file that will be transcribed
            base_model (str) :
                specifies the base_language_model
            language_customization_id (str) :
                ID of the custom acoustic model
            acoustic_customization_id (str) :
                ID of the custom language model.
        """
        # reset the callbacks
        recognize_callbacks.reset()
        assert is_file(audio_path), f"Not a file {audio_path}"

        # Create the stt service and run
        try:
            authenticator = IAMAuthenticator(self.apikey)
            stt = SpeechToTextV1(authenticator=authenticator)
            stt.set_service_url(self.regions[self.region])
        except Exception:
            raise APIKeyError()
        with open(audio_path, "rb") as f:
            # Prepare args
            source = AudioSource(f)
            content_type = self._format_to_content_types[get_extension(audio_path)]
            """ 
            :  confirm current set of key word arguments is okay, 
                       headers and customized weight does not work  """
            kwargs = deepcopy(self.defaults)
            kwargs.update(
                {"inactivity_timeout": -1,
                "audio": source,
                "content_type": content_type,
                "recognize_callback": recognize_callbacks,
                "model": base_model,
                "customization_id": language_customization_id,
                "profanity_filter": setting.profanity_filter,
                "word_confidence":  setting.word_confidence,
                "ssl_verification": setting.ssl_verification,
                "interim_results": setting.interim_results,
                "timestamps": setting.timestamps,
                "smart_formatting": setting.smart_formatting,
                "redaction": setting.redaction,
                "processing_metrics": setting.processing_metrics,
                "audio_metrics": setting.audio_metrics,
                "split_transcript_at_phrase_end": setting.split_transcript_at_phrase_end,
                "inactivity_timeout": setting.inactivity_timeout,
                "keyword_threshold": setting.keyword_threshold,
                "max_alternatives": setting.max_alternatives,
                "processing_metrics_interval": setting.processing_metrics_interval,
                "end_of_phrase_silence_time": setting.end_of_phrase_silence_time,
                "speech_detector_sensitivity": setting.speech_detector_sensitivity,
                "background_audio_supression": setting.background_audio_supression
                }
            )
            stt.recognize_using_websocket(**kwargs)

    def _prepare_utterance(self, audio_name: str, closure: Dict[str, Any]) -> List:
        """
         output the response data from Watson STT, convert the raw data to
        utterance data which is a list of dictionary in the format
        {speaker: , start: , end: , text: }


        Args:
            closure (Dict[str, Any]): contains the result data from Watson

        Returns:
            List:  a list of dictionary that contains the output data
        """
        logger.info(f"preparing utterance for audio {audio_name}")
        utterances = list()
        # Mapping based on (start time, end time)
        data: Dict[Tuple[Any, Any], Dict[str, Any]] = dict()
        # Aggregated data from recognition results
        labels = list()
        timestamps = list()

        # from callback, check that the transcription is successful
        assert not closure["callback_status"]["on_error"]

        # Creating RecognitionResults objects
        for item in closure["results"]["data"]:
            recognition_result = RecognitionResult(item)
            if recognition_result.is_configured():
                labels.extend(recognition_result.get_speaker_labels())
                timestamps.extend(
                    recognition_result.get_timestamps_from_alternatives(
                        only_final=False
                    )
                )
        timestamps = list(chain(*timestamps))

        # Creating the mappings
        for label in labels:  # Label should be a dictionary
            key = (label["start_time"], label["end_time"])
            if key not in data:
                data[key] = {"speaker": label["speaker"]}
            else:
                data[key]["speaker"] = label["speaker"]
        for timestamp in timestamps:
            key = (timestamp[1], timestamp[2])
            if key not in data:
                data[key] = {"utterance": timestamp[0]}
            else:
                data[key]["utterance"] = timestamp[0]

        # Creating utterances
        for times, value in data.items():
            utt = {
                "speaker": value["speaker"],
                "start": times[0],
                "end": times[1],
                "text": value["utterance"],
            }
            utterances.append(utt)
        return utterances

    @staticmethod
    def is_api_valid(apikey: str, url: str) -> bool:
        """
        Determines if given Watson API key is valid

        Args:
            apikey (str) : apikey of which to determine validity
            url (str) : URL to set service url of speech to text service
        """
        try:
            stt = WatsonCore._initialize_stt_service(apikey)
            stt.set_service_url(url)
            stt.list_models()
            return True
        except Exception as e:
            logger.error(e, exc_info=True)
            return False

    @staticmethod
    def _initialize_stt_service(apikey: str) -> SpeechToTextV1:
        """
        Initializes the speech to text services

        Args:
            apikey (str) : API key to pass into the IAMAuthenticator

        Returns:
            Newly initialized speech to text service
        """
        authenticator = IAMAuthenticator(apikey)
        stt = SpeechToTextV1(authenticator=authenticator)
        return stt



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Lakshita Jain
# @Last Modified time: 2023-12-05 16:44:59
from dataclasses import dataclass
from dict_to_dataclass import field_from_dict, DataclassFromDict
import toml
from typing import List, Dict


@dataclass
class DirectoryName(DataclassFromDict):
    temp_extension: str = field_from_dict()
    output_extension: str = field_from_dict()
    hidden_file: str = field_from_dict()
    trans_result: str = field_from_dict()
    temp_result: str = field_from_dict()
    analysis_result: str = field_from_dict()


@dataclass
class Engines(DataclassFromDict):
    engine: str = field_from_dict()
    google_name: str = field_from_dict()
    watson_name: str = field_from_dict()
    whisper_name: str = field_from_dict()
    audio_supported_format: List[str] = field_from_dict()


@dataclass
class Thread(DataclassFromDict):
    transcriber_num_threads: int = field_from_dict()
    payload_num_threads: int = field_from_dict()
    analysis_num_threads: int = field_from_dict()


@dataclass
class ServiceConfig(DataclassFromDict):
    engines: Engines = field_from_dict()
    directory_name: DirectoryName = field_from_dict()
    thread: Thread = field_from_dict()


@dataclass
class ProfileData(DataclassFromDict):
    plugin_setting: Dict[str, Dict[str, List[str]]] = field_from_dict()
    engine_setting_name: str = field_from_dict()


@dataclass
class DefaultSetting(DataclassFromDict):
    profile_name: str = field_from_dict()
    profile_data: dict = field_from_dict()
    engine_name: str = field_from_dict()
    engine_data: dict = field_from_dict()
    profile_data_no_plugin: dict = field_from_dict()


def load_service_config(path: str) -> ServiceConfig:
    d = toml.load(path)
    return ServiceConfig.from_dict(d)


def load_default_setting(path: str) -> DefaultSetting:
    d = toml.load(path)
    return DefaultSetting.from_dict(d)



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Vivian Li 
# @Last Modified time: 2024-05-03 17:14:56

from gailbot.engineManager.engine.engine import Engine
from .googleSetting import GoogleSetting
from ..google.core import GoogleCore
from .AC import GoogleACInterface
from typing import Dict, List
from gailbot.shared.utils.logger import makelogger

logger = makelogger("google engine")


class Google(Engine):
    """
    An Engine that connect to Google Cloud STT, provide function to transcribe
    audio file with Google Cloud STT

    Inheritance:
        Engine
    """
    def __init__(self, setting: GoogleSetting):
        self.setting = setting
        self.ac = GoogleACInterface(api_key_path=self.setting.google_api_key)
        self.core = GoogleCore(self.setting)
        self.transcribe_success = False

    def __repr__(self):
        return self.setting.engine

    def transcribe(self, audio_path: str, workspace: str) -> List[Dict[str, str]]:
        """use Google engine to transcribe the audio file

        Args:
            audio_path (str): path to audio source
            workspace (str): path to a folder that is available for engine to store temporary data

        Raises:
            Err.TranscriptionError

        Returns:
            A list of dictionary that contains the utterance data of the
            audio file, each part of the audio file is stored in the format
            {speaker: , start_time: , end_time: , text: }
        """
        res = self.core.transcribe(audio_path, workspace)
        self.transcribe_success = True
        return res

    def is_file_supported(self, file_path: str) -> bool:
        """
        given a file path, return true if the file format is supported by
        the Google STT engine
        """
        return self.core.is_file_supported(file_path)

    def get_supported_formats(self) -> List[str]:
        return self.core.supported_formats

    def get_engine_name(self) -> str:
        return self.setting.engine

    def was_transcription_successful(self) -> bool:
        return self.transcribe_success

    @staticmethod
    def is_valid_google_api(google_api_key) -> bool:
        """ takes a path to a json file """
        engine = GoogleCore.is_valid_google_api(google_api_key)
        return True if engine else False



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-02-06 15:54:56
# @Description: predefined error message for different reasons when plugin suite is not loaded correctly
from dataclasses import dataclass


@dataclass
class SUITE_REGISTER_MSG:
    """
    a data class that stores the message to report any plugin suite registration error
    """
    INVALID_URL = "The given url is not supported by gailbot\n"
    MODULE_ERROR = "Fail to load import plugin module\n"
    INVALID_INPUT = "The plugin suite source can only be URL, a valid Amazon S3 Bucket name, or path to directory\n"
    TEMPLATE = "ERROR: {cause}\n"
    MISSING = "ERROR: missing file {file}\n"
    FAIL_LOAD_PLUGIN = "ERROR: failed to load plugin {plugin} due to {cause}\n"
    INFO = "INFO: \n"



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-03-29 17:10:22
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-24 19:03:12
from typing import List
from pydantic import BaseModel

from gailbot.setting_interface.formItem import FormItem, FormType



class WhisperXSetting(BaseModel):
    engine: str = "whisperX"
    output_format: str = "all"
    language: str = "en"
    interpolate_method: str = "nearest"
    min_speakers: int = 2
    max_speakers: int = 2
    temperature: float = 1
    condition_on_previous_text: bool = False
    initial_prompt: str = None
    vad_onset: float = 0.5
    vad_offset: float = 0.363
    chunk_size: int =  30
    no_speech_threshold: float = 0.6

    @staticmethod
    def get_setting_config() -> List[FormItem]:
        return [
            FormItem(type=FormType.Selection, name="engine", selection_items=["whisperX"], default_value="whisperX"),

            FormItem(type=FormType.Selection, name="language", selection_items=["af", "am", "ar", "as", "az", "ba", "be", "bg", "bn", "bo", "br", "bs", "ca", "cs", 
            "cy", "da", "de", "el", "en", "es", "et", "eu", "fa", "fi", "fo", "fr", "gl", "gu", 
            "ha", "haw", "he", "hi", "hr", "hu", "hy", "id", "is", "it", "ja", "jw", "ka", 
            "kk", "km", "kn", "ko", "la", "lb", "ln", "lo", "lt", "lv", "mg", "mi", "mk", 
            "ml", "mn", "mr", "ms", "mt", "my", "ne", "nl", "no", "nn", "oc", "pa", "pl", 
            "ps", "pt", "ro", "ru", "sa", "sd", "si", "sk", "sl", "sn", "so", "sq", "sr", 
            "sv", "sw", "ta", "te", "tg", "th", "tl", "tr", "tt", "uk", "ur", "uz", "vi", 
            "yi", "yo", "yue", "zh"], default_value= "en"),
            FormItem(type=FormType.Selection, name="interpolate_method", selection_items=["nearest", "linear", "ignore"], default_value= "nearest"),
            FormItem(type=FormType.Selection, name="min_speakers", selection_items=[1, 2, 3, 4, 5, 6], default_value= 2),
            FormItem(type=FormType.Selection, name="max_speakers", selection_items=[1, 2, 3, 4, 5, 6], default_value= 2),
            FormItem(type=FormType.Selection, name="temperature", selection_items=[0, 0.25, 0.5, 0.75, 1], default_value= 1),

            ### advanced settings:
            FormItem(type=FormType.OnOff, name="condition_on_previous_text", default_value= False),
            FormItem(type=FormType.Text, name="initial_prompt"),
            FormItem(type=FormType.Selection, name="vad_onset", selection_items=[0, 0.1, 0.2, 0.3, 0.4, 0.5], default_value= 0.5),
            FormItem(type=FormType.Selection, name="vad_offset", selection_items=[0, 0.1, 0.2, 0.363], default_value= 0.363),
            FormItem(type=FormType.Selection, name="chunk_size", selection_items=[5, 10, 15, 20, 25, 30], default_value= 30),
            FormItem(type=FormType.Selection, name="no_speech_threshold", selection_items=[0, 0.3, 0.6, 0.9], default_value= 0.6),
            FormItem(
                type=FormType.Number,
                name="num_threads",
                default_value=1,
                min=1,
                max=64
            )
        ]

    


# -*- coding: utf-8 -*-
# @Author: Vivian Li 
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Vivian Li 
# @Last Modified time: 2024-02-06 16:08:36
from typing import List, TextIO

from gailbot.pluginSuiteManager.suite.suite import PluginSuite
from gailbot.pluginSuiteManager.suiteLoader.pluginSuiteLoader import PluginSuiteLoader
from gailbot.pluginSuiteManager.suiteLoader.urlloaders import ZipUrlLoader, \
    S3BucketLoader, UrlLoader


class PluginURLSuiteLoader(PluginSuiteLoader):
    url_loader_classes = [ZipUrlLoader, S3BucketLoader]

    def __init__(self, download_dir: str, suites_dir: str) -> None:
        super().__init__()
        self.download_dir = download_dir
        self.url_loaders: List[UrlLoader] = [
            ZipUrlLoader(download_dir, suites_dir),
            S3BucketLoader(download_dir, suites_dir),
        ]

    @property
    def supported_url_source(self):
        """return a list of supported url downloading source"""
        return ["github", "amazon s3"]

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """
        check if the url string is valid

        Args:
            url (str): a string that represent the url

        Returns:
            bool: true if the string is valid url false otherwise
        """
        for loader in PluginURLSuiteLoader.url_loader_classes:
            if loader.is_supported_url(url):
                return True
        return False

    def load(self, url: str, f: TextIO) -> List[PluginSuite]:
        for loader in self.url_loaders:
            suites = loader.load(url, f)
            if suites:
                return suites
        return []



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-09 10:09:26
# @Description: Implement static functions to transcribe multiple sources in parallel
from concurrent.futures import ThreadPoolExecutor, wait
from dataclasses import dataclass
from typing import List

from gailbot.payload import PayloadObject
from gailbot.shared.utils.logger import makelogger
from gailbot.sourceManager.sourceObject import SourceObject


@dataclass(init=True)
class TranscriptionResult:
    success_output: List[str]
    failure: List[str]
    invalid: List[str]


@dataclass(init=True)
class PayloadTranscriptionResult:
    success: List[PayloadObject]
    failure: List[PayloadObject]


@dataclass(init=True)
class ConvertResult:
    converted: List[PayloadObject]
    invalid: List[SourceObject]
    original_source_ids: List[str]


class GBPipeline:
    logger = makelogger("transcriptionPipeline")

    @staticmethod
    def convert(sources: List[SourceObject]) -> ConvertResult:
        """
        Convert a list of sources to payload objects

        Notes
        ----------
            This is only used for developer to write applications that use gailbot for more fine-grained
            control to the transcription process

        Parameters
        ----------
        sources:
            Each SourceObject in the list sources must be applied with a valid profile setting

        Returns
        -------
            [ConvertResult.converted] stores a list of converted payload
            [ConvertResult.invalid] stores a list of invalid SourceObject
            [ConvertResult.original_source_ids] stores the list of original source id being converted
        """
        converted: List[PayloadObject] = []
        invalid: List[SourceObject] = []
        for s in sources:
            try:
                new_payload = s.convert()
                if not new_payload:
                    invalid.append(s)
                converted.extend(new_payload)
            except Exception as e:
                GBPipeline.logger.error(f"failed to convert {s.source_path}")
                GBPipeline.logger.error(e, exc_info=True)
                invalid.append(s)
        source_ids = [s.name for s in sources]
        return ConvertResult(converted, invalid, source_ids)

    @staticmethod
    def transcribe_payloads(
        payloads: List[PayloadObject],
    ) -> PayloadTranscriptionResult:
        """
        Transcribe a given list of payloads in parallel

        Notes
        ----------
            This is only used for developer to write applications that use gailbot for more fine-grained
            control to the transcription process

        Parameters
        ----------
        payloads:
            A list of valid payload where the transcription will start when payload.start_execute function
            is evoked

        Returns
        -------
            [PayloadTranscriptionResult.success] stores a list of successfully transcribed payload
            [PayloadTranscriptionResult.failure] stores a list of failed payload
        """
        io_thread_limit = 15
        cpu_thread_limit = 1
        success = list()
        failure = list()
        futures = dict()

        cpu_bound = []
        io_bound = []

        for payload in payloads:
            if payload.profile.engine_provider.is_cpu_intensive():
                cpu_bound.append(payload)
            else:
                io_bound.append(payload)

        # use separate threadpool executor for cpu and io bound engine
        if io_bound:
            io_executor = ThreadPoolExecutor(
                max_workers=min(len(io_bound), io_thread_limit)
            )
            for payload in io_bound:
                futures[payload] = io_executor.submit(payload.start_execute)

        if cpu_bound:
            cpu_executor = ThreadPoolExecutor(max_workers=cpu_thread_limit)
            for payload in cpu_bound:
                futures[payload] = cpu_executor.submit(payload.start_execute)

        wait(list(futures.values()))
        for payload, future in futures.items():
            if future.result():
                success.append(payload)
            else:
                failure.append(payload)

        return PayloadTranscriptionResult(success=success, failure=failure)

    @staticmethod
    def transcribe(sources: List[SourceObject]) -> TranscriptionResult:
        """
        Transcribe the list of sources

        Notes
        ----------
            This is the wrapper function for convert and transcribe_payloads used by gailbot api

        Parameters
        ----------
        sources:
            Each SourceObject in the list sources must be applied with a valid profile setting

        Returns
        -------
            [TranscriptionResult.success] stores the output path to the successfully transcribed payloads
            [TranscriptionResult.fail] stores the names of failure sources
            [TranscriptionResult.invalid] stores the names of invalid sources

        """
        payload_result = GBPipeline.convert(sources)
        transcribe_result = GBPipeline.transcribe_payloads(payload_result.converted)
        return TranscriptionResult(
            [s.output_dir for s in transcribe_result.success],
            [f.name for f in transcribe_result.failure],
            [i.name for i in payload_result.invalid],
        )



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-07-24 14:34:38
import toml


def load_valid_structure(path):
    return toml.load(path)



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-31 16:21:27
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-08-01 05:37:41
from .analysis import AnalysisResult, AnalysisResultDict
from .format import FormatResult, FormatResultDict
from .sttTranscribe import UttDict, SttTranscribeResult
from .resultInterface import ResultInterface
from .processingStatus import ProcessingStats



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-08 12:26:09
# @Last Modified by:   Muhammad Umair
# @Last Modified time: 2023-01-09 11:51:25


from .watson import Watson
from .lm import WatsonLMInterface
from .am import WatsonAMInterface




# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-04-24 19:17:48
# @Last Modified by:   Sophie Clemens, Dan Bergen, Eva Caro
# @Last Modified time: 2024-05-28 14:34:00
from typing import List
from pydantic import BaseModel
from typing import Optional
from gailbot.setting_interface.formItem import FormItem, FormType
from gailbot.setting_interface.formItem import FormItem
from gailbot.engineManager.engine.watson.lm import WatsonLMInterface


class WatsonSetting(BaseModel):
    engine: str = "watson"
    apikey: str
    region: str
    base_model: str = "en-US_BroadbandModel"
    custom_model: str = ""

    ssl_verification: bool = True
    interim_results: bool = False
    word_confidence: bool = False
    timestamps: bool = True
    profanity_filter: bool = False
    smart_formatting: bool = True
    redaction: bool = False
    processing_metrics: bool = False
    audio_metrics: bool = False
    split_transcript_at_phrase_end: bool = False

    language_customization_id: Optional[str] = None
    acoustic_customization_id: Optional[str] = None

    inactivity_timeout: int = 30
    keyword_threshold: float = 0.5
    max_alternatives: int = 1
    processing_metrics_interval: float = 0.5
    end_of_phrase_silence_time: float = 1.0
    speech_detector_sensitivity: float = 0.5
    background_audio_supression: float = 0.5

    @staticmethod
    def get_setting_config() -> List[FormItem]:

        return [ 
            FormItem(type=FormType.Selection, name="engine", selection_items=["watson"]),
            FormItem(type=FormType.Text, name="apikey"),
            FormItem(type=FormType.Selection, name="region", selection_items=["dallas", "washington", "frankfurt", "sydney", "tokyo", "london", "seoul"]),
            FormItem(type=FormType.Selection, name="base_model", selection_items=['en-US_BroadbandModel', 'pt-BR_BroadbandModel', 'ja-JP_BroadbandModel', 'es-CL_BroadbandModel', 'es-MX_BroadbandModel', 'zh-CN_BroadbandModel', 'es-ES_BroadbandModel', 'es-CO_BroadbandModel', 'en-US_Telephony', 'es-AR_BroadbandModel', 'es-PE_BroadbandModel', 'it-IT_Telephony', 'en-GB_BroadbandModel', 'nl-NL_NarrowbandModel', 'ko-KR_Multimedia', 'en-GB_NarrowbandModel', 'hi-IN_Telephony', 'sv-SE_Telephony', 'ar-MS_BroadbandModel', 'ja-JP_Multimedia', 'nl-NL_Multimedia', 'nl-NL_Telephony', 'pt-BR_Telephony', 'it-IT_Multimedia', 'fr-FR_BroadbandModel', 'de-DE_NarrowbandModel', 'it-IT_BroadbandModel', 'ko-KR_NarrowbandModel', 'en-GB_Telephony', 'en-US_Multimedia', 'es-PE_NarrowbandModel', 'de-DE_BroadbandModel', 'en-AU_Multimedia', 'es-ES_Telephony', 'en-AU_BroadbandModel', 'fr-CA_NarrowbandModel', 'de-DE_Telephony', 'es-LA_Telephony', 'ja-JP_NarrowbandModel', 'de-DE_Multimedia', 'nl-NL_BroadbandModel', 'es-CO_NarrowbandModel', 'en-GB_Multimedia', 'ko-KR_BroadbandModel', 'en-US_ShortForm_NarrowbandModel', 'es-MX_NarrowbandModel', 'fr-CA_BroadbandModel', 'cs-CZ_Telephony', 'es-ES_NarrowbandModel', 'es-CL_NarrowbandModel', 'fr-CA_Multimedia', 'fr-FR_Telephony', 'fr-FR_NarrowbandModel', 'en-AU_NarrowbandModel', 'en-US_NarrowbandModel', 'en-WW_Medical_Telephony', 'es-AR_NarrowbandModel', 'zh-CN_Telephony', 'it-IT_NarrowbandModel', 'pt-BR_NarrowbandModel', 'en-AU_Telephony', 'ko-KR_Telephony', 'fr-FR_Multimedia', 'es-ES_Multimedia', 'en-IN_Telephony', 'ja-JP_Telephony', 'zh-CN_NarrowbandModel', 'ar-MS_Telephony', 'fr-CA_Telephony', 'pt-BR_Multimedia', 'nl-BE_Telephony'], default_value= "en-US_BroadbandModel"),
            FormItem(type=FormType.Selection, name="custom_model", selection_items=["None"]),

            # Booleans
            FormItem(type=FormType.OnOff, name="ssl_verification", default_value= False),
            FormItem(type=FormType.OnOff, name="interim_results", default_value= False),
            FormItem(type=FormType.OnOff, name="word_confidence", default_value= False),
            FormItem(type=FormType.OnOff, name="timestamps", default_value= False),
            FormItem(type=FormType.OnOff, name="profanity_filter", default_value= True),
            FormItem(type=FormType.OnOff, name="smart_formatting", default_value= False),
            FormItem(type=FormType.OnOff, name="redaction", default_value= False),
            FormItem(type=FormType.OnOff, name="processing_metrics", default_value= False),
            FormItem(type=FormType.OnOff, name="audio_metrics", default_value= False),
            FormItem(type=FormType.OnOff, name="split_transcript_at_phrase_end", default_value= True),
            # Select
            FormItem(type=FormType.Selection, name="inactivity_timeout", selection_items=[-1, 0, 1, 5, 10, 30], default_value= 30),
            FormItem(type=FormType.Selection, name="keyword_threshold", selection_items=[0.1, 0.5, 0.8, 0.9], default_value= None),
            FormItem(type=FormType.Selection, name="max_alternatives", selection_items=[1, 2, 3], default_value= 1),
            FormItem(type=FormType.Selection, name="processing_metrics_interval", selection_items=[0.5, 1.0, 2.0], default_value= None),
            FormItem(type=FormType.Selection, name="end_of_phrase_silence_time", selection_items=[0.5, 0.8, 1.0], default_value= 0.8),
            FormItem(type=FormType.Selection, name="speech_detector_sensitivity", selection_items=[0.3, 0.5, 0.7], default_value= 0.5),
            FormItem(type=FormType.Selection, name="background_audio_supression", selection_items=[0.0, 0.1, 0.2]),
    
        ]
    


# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2021-12-02 13:13:08
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-11 00:20:37
# Standard imports
from typing import List, Dict

# Local imports
# Third party imports
from copy import deepcopy
from ibm_watson.websocket import RecognizeCallback
from gailbot.shared.utils.logger import makelogger

logger = makelogger("callback")


class WatsonException(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)
        self.msg = msg


class CustomWatsonCallbacks(RecognizeCallback):
    """
    Extends the watson callback class to allow custom callbacks to be executed
    when an event occurs through the lifecycle of the websocket connection.

    Inherits:
        (RecognizeCallback)
    """

    def __init__(self) -> None:
        super().__init__()
        self.closure = self._init_closure()

    def reset(self) -> None:
        logger.info("Watson: reset recognize callback")
        self.closure = self._init_closure()

    def get_results(self) -> Dict:
        logger.info("Watson: on get result")
        return deepcopy(self.closure)

    def on_transcription(self, transcript: List) -> None:
        """
        Called after the service returns the final result for the transcription.
        """
        logger.info("Watson: on transcription")
        try:
            closure = self.closure
            closure["callback_status"]["on_transcription"] = True
            closure["results"]["transcript"].append(transcript)
        except Exception as e:
            logger.error(e, exc_info=e)

    def on_connected(self) -> None:
        """
        Called when a Websocket connection was made
        """
        logger.info("Watson: connected to watson")
        try:
            closure = self.closure
            closure["callback_status"]["on_connected"] = True
        except Exception as e:
            logger.error(e, exc_info=e)

    def on_error(self, error: str) -> None:
        """
        Called when there is an error in the Websocket connection.
        """
        logger.error(f"Watson: get error {error}")
        closure = self.closure
        closure["callback_status"]["on_error"] = True
        closure["results"]["error"] = error
        raise WatsonException(error)

    def on_inactivity_timeout(self, error: str) -> None:
        """
        Called when there is an inactivity timeout.
        """
        logger.warning("Watson: inactivity time out")
        try:
            closure = self.closure
            closure["callback_status"]["on_inactivity_timeout"] = True
            closure["results"]["error"] = error
        except Exception as e:
            logger.error(f"timeout error {e}")

    def on_listening(self) -> None:
        """
        Called when the service is listening for audio.
        """
        logger.info("Watson: is listening")
        try:
            closure = self.closure
            closure["callback_status"]["on_listening"] = True
        except Exception as e:
            logger.error(f"on listening error {e}")

    def on_hypothesis(self, hypothesis: str) -> None:
        """
        Called when an interim result is received.
        """
        logger.info(f"Watson: on hypothesis {hypothesis}")
        try:
            closure = self.closure
            closure["callback_status"]["on_hypothesis"] = True
        except Exception as e:
            logger.error(f"on hypothesis error {e}")

    def on_data(self, data: Dict) -> None:
        """
        Called when the service returns results. The data is returned unparsed.
        """
        logger.info(f"Watson: returned the results")
        try:
            closure = self.closure
            closure["callback_status"]["on_data"] = True
            closure["results"]["data"].append(data)
        except Exception as e:
            logger.error(f"on data error {e}", exc_info=True)

    def on_close(self) -> None:
        """
        Called when the Websocket connection is closed
        """
        logger.info("Watson: on close")
        try:
            closure = self.closure
            closure["callback_status"]["on_close"] = True
        except Exception as e:
            logger.error(f"on close error {e}", exc_info=True)

    def _init_closure(self) -> Dict:
        return {
            "callback_status": {
                "on_transcription": False,
                "on_connected": False,
                "on_error": False,
                "on_inactivity_timeout": False,
                "on_listening": False,
                "on_hypothesis": False,
                "on_data": False,
                "on_close": False,
            },
            "results": {
                "error": None,
                "transcript": list(),
                "hypothesis": list(),
                "data": list(),
            },
        }



import os
from typing import Dict, Optional
from gailbot.engineManager.engine.engine import Engine
from gailbot.engineManager.engine.engineProvider import EngineProvider
from .deepgram import Deepgram
from .deepgramSetting import DeepgramSetting
from gailbot.shared.utils.general import make_dir, copy
from gailbot.shared.utils.logger import makelogger
from gailbot.workspace import WorkspaceManager

logger = makelogger("deepgram provider")

class DeepgramProvider(EngineProvider):
    workspace = os.path.join(WorkspaceManager.engine_src, "deepgram/api")

    def is_cpu_intensive(self) -> bool:
        return False

    @staticmethod
    def load(name: str, data: Dict[str, str]) -> Optional['DeepgramProvider']:
        try:
            prov = DeepgramProvider(name, data)
            assert prov.data.engine.lower() == "deepgram"
            return prov
        except Exception as e:
            logger.warning(e)
            return None

    def __init__(self, name: str, data: Dict[str, str]):
        if not os.path.isdir(self.workspace):
            make_dir(self.workspace)
        self.name = name
        self.data: DeepgramSetting = DeepgramSetting(**data)
        assert Deepgram.is_valid_deepgram_api(self.data.deepgram_api_key)

    def make_engine(self) -> Engine:
        return Deepgram(self.data)



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-16 11:07:12
# @Last Modified by:   Joanne Fan
# @Last Modified time: 2024-05-16 16:07:46

from enum import IntEnum

class WatsonReturnCodes(IntEnum):
    """
    Return codes from Watson.
    """
    OK = 200
    CREATED = 201
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    NOT_FOUND = 404
    NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    CONFLICT = 409
    UNSUPPORTED = 415
    SERVER_ERROR = 500
    SERVICE_UNAVAILABLE = 503



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-08 12:43:29
# @Last Modified by:   Lakshita Jain
# @Last Modified time: 2023-11-30 16:53:18

from typing import Dict, List

from gailbot.engineManager.engine.engine import Engine
from gailbot.engineManager.engine.whisper.core import WhisperCore
from gailbot.shared.utils.general import get_extension
from gailbot.shared.utils.logger import makelogger
from .whisperSetting import WhisperSetting

logger = makelogger("Whisper Engine")


class Whisper(Engine):
    def __init__(self, setting: WhisperSetting):
        self.setting = setting
        self.core = WhisperCore()
        self._successful = False

    def __str__(self):
        """
        Returns the name of the function
        """
        return self.setting.engine_name

    def __repr__(self):
        """
        Returns all the configurations and additional metadata
        """
        return self.core.__repr__()

    def transcribe(
            self,
            audio_path, payload_workspace
    ) -> List[Dict]:
        results = self.core.transcribe(audio_path, self.setting.language, self.setting.detect_speakers)
        self._successful = True
        return results

    def was_transcription_successful(self) -> bool:
        """
        Return true if transcription was successful, false otherwise.
        """
        return self._successful

    def get_engine_name(self) -> str:
        """
        Obtain the name of the current engine.

        Returns:
            (str): Name of the engine.
        """
        return self.setting.engine

    def get_supported_formats(self) -> List[str]:
        """
        Obtain a list of audio file formats that are supported.

        Returns:
            (List[str]): Supported audio file formats.
        """
        return self.core.get_supported_formats()

    def is_file_supported(self, file_path: str) -> bool:
        """
        Determine if the given file is supported by the engine.

        Args:
            file_path (str)

        Returns:
            (bool): True if file is supported. False otherwise.
        """
        return get_extension(file_path) in self.get_supported_formats()

    def get_available_models(self) -> List[str]:
        """
        Return the list of available models
        """
        return self.core.get_available_models()



# -*- coding: utf-8 -*-
# @Author: Dan Bergen
# @Date:   2023-06-06 14:36:00
# @Last Modified by:   Your Name
# @Last Modified time: Your Time

# EXCELLENT RESOURCE FOR UNDERSTANDING: https://cloud.google.com/speech-to-text/docs/adaptation-model#improve_transcription_results_using_a_customclass

# Standard library imports
from typing import Any, List, Dict, Tuple, Callable

# Local imports

# Third party imports
from google.cloud import speech_v1p1beta1 as speech
from google.api_core.exceptions import GoogleAPIError
from google.oauth2 import service_account
from google.protobuf import field_mask_pb2

import logging, os

logger = logging.getLogger(__name__)

class GoogleACInterface:
    def __init__(self, api_key_path: str = None, api_key_str: str = None):
        """
        Initialize the GoogleCCInterface with API credentials.

        Args:
            api_key_path (str): Path to the JSON file containing the service account key.

        Raises:
            Exception: If the API key is invalid.
        """
        # Set up the Google Cloud credentials
       

        try:
            if (api_key_str == None) :
                self.credentials = service_account.Credentials.from_service_account_file(api_key_path)
            else: 
                self.credentials = service_account.Credentials.from_service_account_info(api_key_str)
            self.project_id = self.credentials.project_id
            print(self.project_id)
            self.client = speech.AdaptationClient(credentials=self.credentials)
        except:
            raise Exception("Connect to STT failed")

    def create_custom_class(self, custom_class_id: str, phrases: List[str]) -> Dict[str, Any]:
        """
        Create a new custom class with the given phrases.

        Args:
            custom_class_id (str): ID for the custom class.
            phrases (List[str]): List of phrases to be included in the custom class.

        Returns:
            Dict[str, Any]: The created custom class resource.

        Raises:
            GoogleAPIError: If the request fails.
        """
        parent = f"projects/{self.project_id}/locations/global"
        custom_class = speech.CustomClass(
            items=[speech.CustomClass.ClassItem(value=phrase) for phrase in phrases]
        )
        response = self._execute_google_method(self.client.create_custom_class, parent=parent, custom_class=custom_class, custom_class_id=custom_class_id)
        logger.info(f"Created Custom Class: {response.name}")
        return response

    def get_custom_class(self, custom_class_id: str) -> Dict[str, Any]:
        """
        Retrieve a custom class by its ID.

        Args:
            custom_class_id (str): ID of the custom class to retrieve.

        Returns:
            Dict[str, Any]: The retrieved custom class resource.

        Raises:
            GoogleAPIError: If the request fails.
        """
        name = f"projects/{self.project_id}/locations/global/customClasses/{custom_class_id}"
        custom_class = self._execute_google_method(self.client.get_custom_class, name=name)
        return custom_class

    def delete_custom_class(self, custom_class_id: str) -> None:
        """
        Delete a custom class by its ID.

        Args:
            custom_class_id (str): ID of the custom class to delete.

        Returns:
            None

        Raises:
            GoogleAPIError: If the request fails.
        """
        name = f"projects/{self.project_id}/locations/global/customClasses/{custom_class_id}"
        self._execute_google_method(self.client.delete_custom_class, name=name)
        logger.info(f"Deleted Custom Class: {name}")

    def update_custom_class(self, custom_class_id: str, phrases: List[str]) -> Dict[str, Any]:
        """
        Update an existing custom class with new phrases.

        Args:
            custom_class_id (str): ID of the custom class to update.
            phrases (List[str]): New list of phrases to include in the custom class.

        Returns:
            Dict[str, Any]: The updated custom class resource.

        Raises:
            GoogleAPIError: If the request fails.
        """
        name = f"projects/{self.project_id}/locations/global/customClasses/{custom_class_id}"
        custom_class = self.get_custom_class(custom_class_id)
        custom_class.items = [speech.CustomClass.ClassItem(value=phrase) for phrase in phrases]
        update_mask = field_mask_pb2.FieldMask(paths=['items'])
        response = self._execute_google_method(self.client.update_custom_class, custom_class=custom_class, update_mask=update_mask)
        logger.info(f"Updated Custom Class: {response.name}")
        return response

    def list_custom_classes(self) -> List[Dict[str, Any]]:
        """
        List all custom classes in the project.

        Returns:
            List[Dict[str, Any]]: A list of all custom class resources.

        Raises:
            GoogleAPIError: If the request fails.
        """
        parent = f"projects/{self.project_id}/locations/global"
        custom_classes = self._execute_google_method(self.client.list_custom_classes, parent=parent)
        return list(custom_classes)
    
    def create_phrase_set(self, phrase_set_id: str, phrases: List[Tuple[str, int]]) -> Dict[str, Any]:
        """
        Create a new phrase set with the given phrases.

        Args:
            phrase_set_id (str): ID for the phrase set.
            phrases (List[Tuple[str, int]]): List of tuples where each tuple contains a phrase and a boost value.

        Returns:
            Dict[str, Any]: The created phrase set resource.

        Raises:
            GoogleAPIError: If the request fails.
        """
        parent = f"projects/{self.project_id}/locations/global"
        phrase_set = speech.PhraseSet(
            phrases=[speech.PhraseSet.Phrase(value=phrase, boost=boost) for phrase, boost in phrases]
        )
        response = self._execute_google_method(self.client.create_phrase_set, parent=parent, phrase_set=phrase_set, phrase_set_id=phrase_set_id)
        logger.info(f"Created Phrase Set: {response.name}")
        return response


    def get_phrase_set(self, phrase_set_id: str) -> Dict[str, Any]:
        """
        Retrieve a phrase set by its ID.

        Args:
            phrase_set_id (str): ID of the phrase set to retrieve.

        Returns:
            Dict[str, Any]: The retrieved phrase set resource.

        Raises:
            GoogleAPIError: If the request fails.
        """
        name = f"projects/{self.project_id}/locations/global/phraseSets/{phrase_set_id}"
        phrase_set = self._execute_google_method(self.client.get_phrase_set, name=name)
        return phrase_set

    def delete_phrase_set(self, phrase_set_id: str) -> None:
        """
        Delete a phrase set by its ID.

        Args:
            phrase_set_id (str): ID of the phrase set to delete.

        Returns:
            None

        Raises:
            GoogleAPIError: If the request fails.
        """
        name = f"projects/{self.project_id}/locations/global/phraseSets/{phrase_set_id}"
        self._execute_google_method(self.client.delete_phrase_set, name=name)
        logger.info(f"Deleted Phrase Set: {name}")

    def update_phrase_set(self, phrase_set_id: str, phrases: List[Tuple[str, int]]) -> Dict[str, Any]:
        """
        Update an existing phrase set with new phrases.

        Args:
            phrase_set_id (str): ID of the phrase set to update.
            phrases (List[Tuple[str, int]]): New list of tuples where each tuple contains a phrase and a boost value.

        Returns:
            Dict[str, Any]: The updated phrase set resource.

        Raises:
            GoogleAPIError: If the request fails.
        """
        name = f"projects/{self.project_id}/locations/global/phraseSets/{phrase_set_id}"
        phrase_set = self.get_phrase_set(phrase_set_id)
        phrase_set.phrases = [speech.PhraseSet.Phrase(value=phrase, boost=boost) for phrase, boost in phrases]
        update_mask = field_mask_pb2.FieldMask(paths=['phrases'])       
        response = self._execute_google_method(self.client.update_phrase_set, phrase_set=phrase_set, update_mask=update_mask)
        logger.info(f"Updated Phrase Set: {response.name}")
        return response

    def list_phrase_sets(self) -> List[Dict[str, Any]]:
        """
        List all phrase sets in the project.

        Returns:
            List[Dict[str, Any]]: A list of all phrase set resources.

        Raises:
            GoogleAPIError: If the request fails.
        """
        parent = f"projects/{self.project_id}/locations/global"
        phrase_sets = self._execute_google_method(self.client.list_phrase_set, parent=parent)
        return list(phrase_sets)

    def _execute_google_method(
        self,
        method: Callable,
        *args: Any,
        **kwargs: Any
    ) -> Any:
        """
        Execute a Google API method and handle exceptions.

        Args: n
            method (Callable): Google API method to execute.
            args (Any): Positional arguments to pass to the method.
            kwargs (Any): Keyword arguments to pass to the method.

        Returns:
            Any: The result of the method call if successful.

        Raises:
            GoogleAPIError: If the method call fails.
        """
        try:
            response = method(*args, **kwargs)
            return response
        except GoogleAPIError as e:
            logger.error(f"Google API method execution failed: {e}", exc_info=e)
            raise e


# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-02-23 18:03:11
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-02-24 09:55:07

from .profileManager import ProfileManager, ProfileObject



from abc import ABC


class GBException(ABC, Exception):
    """
    abstract class for GB exception
    """

    def __init__(self):
        self.code = 0
        self.error_msg: str = ""

    def __str__(self):
        return f"ERROR: {self.code} {self.error_msg}"



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Joanne Fan
# @Last Modified time: 2024-04-01 02:20:18
from typing import Dict, List

from gailbot.shared.exception.serviceException import SourceNotFound
from gailbot.shared.utils.general import get_name
from gailbot.sourceManager.sourceObject import SourceObject, Watcher
from gailbot.payload.payloadConverter import (ConversationDirConverter,
                                              MediaConverter,
                                              TranscribedDirConverter,
                                              MixedDirectoryConverter,
                                              ConverterType)


class SourceManager:
    def __init__(self):
        self._sources: Dict[str, SourceObject] = dict()
        self._converters = {
            ConverterType.ConvDir: ConversationDirConverter,
            ConverterType.SingleMedia: MediaConverter,
            ConverterType.TranscribedResult: TranscribedDirConverter,
            ConverterType.MixedDirectory: MixedDirectoryConverter
        }

    def available_converters(self) -> List[ConverterType]:
        return list(self._converters.keys())

    def un_transcribed_sources(self) -> Dict[str, SourceObject]:
        return self._sources

    def add_source(self, source: str, output: str) -> str:
        source_id = get_name(source)
        if source_id in self._sources:
            counter = 1
            temp = source_id + f"({counter})"
            while temp in self._sources:
                counter += 1
                temp = source_id + f"({counter})"
            source_id = temp
        self._sources[source_id] = SourceObject(source_id, source, output)
        return source_id

    def is_source(self, source_id: str):
        return source_id in self._sources

    def get_source_output(self, source_id: str):
        if not self.is_source(source_id):
            raise SourceNotFound(source_id)
        return self._sources[source_id].output_dir

    def remove_source(self, source_id: str):
        if source_id not in self._sources:
            raise SourceNotFound(source_id)
        else:
            self._sources.pop(source_id)

    def apply_profile_to_source(self, source_id: str, profile: str):
        if source_id not in self._sources:
            raise SourceNotFound(source_id)
        else:
            self._sources[source_id].link_profile(profile)

    def get_source_profile(self, source_id: str) -> str:
        if not self.is_source(source_id):
            raise SourceNotFound(source_id)
        else:
            return self._sources[source_id].profile_name

    def get_source(self, source_id) -> SourceObject:
        if not self.is_source(source_id):
            raise SourceNotFound(source_id)
        else:
            return self._sources[source_id]

    def profile_used_by_source(self, profile_name: str) -> bool:
        """
        return true if [profile_name] is used by any source, false otherwise
        """
        for source in self._sources.values():
            if source.profile_obj == profile_name:
                return True
        else:
            return False

    def add_progress_watcher(self, source_id: str, watcher: Watcher):
        if not self.is_source(source_id):
            raise SourceNotFound(source_id)
        else:
            self._sources[source_id].add_watcher(watcher)

    def change_source_converter(self, source_id: str, converter: ConverterType):
        converter = self._converters[converter]
        return self._sources[source_id].change_converter(converter)

    def get_source_converter(self, source_id: str) -> ConverterType:
        return self.get_source(source_id).converter.converter_type

    def compatible_converters(self, source_id: str) -> List[ConverterType]:
        compatible = self._sources[source_id].compatible_converters()
        return [key for key, value in self._converters.items() if value in compatible]


import os
from dataclasses import dataclass

from gailbot.shared.utils.general import make_dir
from gailbot.workspace import WorkspaceManager


@dataclass
class TemporaryFolder:
    def __init__(self, name: str):
        self.root = os.path.join(WorkspaceManager.temporary_ws, name)
        self.data_copy = os.path.join(self.root, "data_copy")
        self.analysis_ws = os.path.join(self.root, "analysis_ws")
        self.transcribe_ws = os.path.join(self.root, "transcribe_ws")
        self.format_ws = os.path.join(self.root, "format_ws")
        for directory in [self.root, self.data_copy, self.analysis_ws, self.transcribe_ws, self.format_ws]:
            if not os.path.isdir(directory):
                make_dir(directory, overwrite=True)


@dataclass
class OutputFolder:
    GB_RESULT_SIGNATURE = ".gailbot"
    MERGED_FILE_NAME = "merged"
    MERGED_FILE = "merged.wav"

    def __init__(self, root):
        self.root = root
        self.raw = os.path.join(self.root, "Raw")
        self.analysis = os.path.join(self.root, "Analysis")
        self.transcribe = os.path.join(self.raw, "Transcript")
        self.media = os.path.join(self.raw, "Media")
        self.about = os.path.join(self.root, "About")
        for directory in [self.root, self.raw, self.analysis, self.transcribe, self.media, self.about]:
            make_dir(directory, overwrite=True)

    @staticmethod
    def get_transcribe_dir(root: str):
        return os.path.join(root, "Raw/Transcript")



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-07-04 15:35:16
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-02-06 15:43:20
# @Description: Manage backend configuration files

import os

PROJECT_ROOT = os.path.dirname(os.path.dirname(__file__))



import os
SUITE_MANAGER_TEST_PATH = os.path.dirname(__file__)
SUITE_TEST_WORKSPACE_PATH = os.path.join(SUITE_MANAGER_TEST_PATH, "suite_workspace")
SUITE_TEST_SRC_PATH = os.path.join(SUITE_MANAGER_TEST_PATH, "suite_src")



# -*- coding: utf-8 -*-
# @Author: Vivian Li, Siara Small
# @Date:   2023-01-30 16:00
# @Last Modified by:   Riddhi Sahni
# @Last Modified time: 2024-04-26 15:02:58

# stdlib import
import os
import io
from copy import deepcopy
from typing import Dict, List

# internal import
from gailbot.shared.utils.general import (
    get_extension,
    is_directory,
    make_dir,
    get_name,
)
from gailbot.engineManager.engine.google.googleSetting import GoogleSetting
from gailbot.shared.utils.download import is_internet_connected
from gailbot.shared.utils.logger import makelogger
from gailbot.shared.exception.transcribeException import (
    InternetConnectionError,
    TranscriptionError,
)
from gailbot.configs import google_config_loader
from gailbot.shared.utils.media import MediaHandler

# external import
from google.cloud.speech_v1p1beta1.types import cloud_speech
from google.cloud import speech_v1p1beta1 as speech
from gailbot.workspace import WorkspaceManager

logger = makelogger("google")
GOOGLE_CONFIG = google_config_loader()


class GoogleCore:
    """
    Implement shared functionalities to transcribe an audio file through
    google STT engine
    """

    ENCODING_TABLE = {
        "wav": speech.RecognitionConfig.AudioEncoding.LINEAR16,
        "mp3": speech.RecognitionConfig.AudioEncoding.MP3,
        "opus": speech.RecognitionConfig.AudioEncoding.OGG_OPUS,
        "flac": speech.RecognitionConfig.AudioEncoding.FLAC,
        "amr": speech.RecognitionConfig.AudioEncoding.AMR,
    }

    def __init__(self, google_setting: GoogleSetting) -> None:
        """
        initialize an instance of google engine

        Args:
            google_api_key_config (str): path to a json file that stores
            google cloud speech to text api key
        """
        self._init_status()
        self.setting = google_setting
        client = GoogleCore.is_valid_google_api(self.setting.google_api_key)
        self.workspace = os.path.join(WorkspaceManager.engine_src, "google")
        if not os.path.isdir(self.workspace):
            make_dir(self.workspace)
        assert client
        self.connected = True
        self.current_chunk_duration = GOOGLE_CONFIG.maximum_duration

    @staticmethod
    def is_valid_google_api(google_api_key_config: str):
        """Given a path to a json file that stores the google api key
            return the google client if the api is valid, else return false

        Args:
            google_api_key_config (str): a path to a file that stores the google api
                                 key

        Returns:
            Union(SpeechCient, bool): if the api key is valid, return the google
                                      speech client, else return false
        """
        try:
            assert (
                get_extension(google_api_key_config) == "json"
            ), "The google api key file is in wrong format"
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = google_api_key_config
            speech.SpeechClient()
            return True
        except Exception as e:
            logger.error("failed to connect to google", exc_info=e)
            return False

    @property
    def supported_formats(self) -> List[str]:
        """
        Access the supported audio formats

        Returns:
            List[str] : list of supported formats
        """
        return list(self.ENCODING_TABLE.keys())

    def is_file_supported(self, file: str) -> bool:
        """
        Determines if a given file is supported

        Args:
            file (str) : file name to check if supported

        Returns: True if the file is supported, false if not.
        """
        return get_extension(file) in self.supported_formats

    def transcribe(
        self, audio_path: str, payload_workspace: str
    ) -> List[Dict[str, str]]:
        """
        Transcribes the provided audio stream using a websocket connections.
        And output the result to given output directory

        Parameters
        ----------
        audio_path
        payload_workspace


        Return
        -----------
            A list of dictionary that contains the utterance data of the
            audio file, each part of the audio file is stored in the format
            {speaker: , start_time: , end: , text: }

        """
        logger.info("starting transcribing audio using google cloud STT")
        if not is_internet_connected():
            raise InternetConnectionError()
        # get the info of audio source and preprocess the audio if needed
        mediaHandler = MediaHandler()
        stream = mediaHandler.read_file(audio_path)
        info = mediaHandler.info(stream)

        audio_duration = info["duration_seconds"]
        audio_channels = info["channels"]
        audio_format = info["format"]

        #  convert stereo to mono
        if audio_channels > 1:
            logger.info("detected stereo audio file, converting to mono file")
            l_stream, _ = mediaHandler.stereo_to_mono(stream)
            audio_path = mediaHandler.write_stream(
                l_stream,
                out_dir=payload_workspace,
                name=get_name(audio_path),
                format=info["format"],
            )

        # convert the wav file to all bit 16 bit
        if audio_format.lower() == "wav":
            logger.info("detected wav file, converting to wav 16 bit header file")
            wav_copy_ws = os.path.join(
                payload_workspace, f"{get_name(audio_path)}_wav_16bit"
            )
            if not is_directory(wav_copy_ws):
                make_dir(wav_copy_ws)
            output_for16 = os.path.join(
                wav_copy_ws, f"{get_name(audio_path)}_16bit.wav"
            )
            audio_path = MediaHandler.convert_to_16bit_wav(audio_path, output_for16)

        # chunk the audio lager than 60 seconds
        if audio_duration >= GOOGLE_CONFIG.maximum_duration:
            logger.info("audio length exceeds maximum limit")
            # get the duration each chunk
            chunk_duration = GoogleCore._get_chunk_duration(audio_path, audio_duration)
            self.current_chunk_duration = chunk_duration
            # chunk the file
            audio_list = MediaHandler.chunk_audio_to_outpath(
                audio_path, payload_workspace, chunk_duration
            )
        else:
            audio_list = [audio_path]

        try:
            return self._transcribe_list_file(audio_list, payload_workspace)
        except Exception as e:
            logger.error(e, exc_info=e)
            raise TranscriptionError(e)

    def _run_engine(self, audio_path: str, workspace) -> cloud_speech.RecognizeResponse:
        """
        run the Google STT engine to transcribe the file

        Args:
            audio_path (str):
                path to audio file that will be transcribed

        Return (cloud_speech.RecognizeResponse):
            return the response data from Google STT
        """
        logger.info(f"running google engine on file {audio_path}")
        try:
            # initialize client
            client = speech.SpeechClient()

            # preprocessing the file
            format = get_extension(audio_path).lower()
            encoding = self.ENCODING_TABLE[format]
            kwargs = deepcopy(GOOGLE_CONFIG.defaults)
            kwargs.update(
                {"language_code": self.setting.language_code})
            kwargs.update(
                {"enable_separate_recognition_per_channel": self.setting.enable_separate_recognition_per_channel})
            kwargs.update(
                {"enable_word_time_offsets": self.setting.enable_word_time_offsets})
            kwargs.update(
                {"enable_word_confidence": self.setting.enable_word_confidence})
            kwargs.update(
                {"enable_automatic_punctuation": self.setting.enable_automatic_punctuation})
            kwargs.update(
                {"enable_spoken_punctuation": self.setting.enable_spoken_punctuation})
            kwargs.update(
                {"enable_spoken_emojis": self.setting.enable_spoken_emojis})
            kwargs.update(
                {"use_enhanced": self.setting.use_enhanced})
            kwargs.update(
                {"model": self.setting.model})
            kwargs.update(
                {"sample_rate_hertz": self.setting.sample_rate_hertz})
            kwargs.update(
                {"profanity_filter": self.setting.profanity_filter})
            
            # add additional key word arguments for file that is no in wav or flac format
            #
            if not (format == "wav" or format == "flac"):
                kwargs.update(
                    {
                        "encoding": encoding,
                        "sample_rate_hertz": 16000,
                    }
                )

            if (len(self.setting.phrase_sets) > 0):
                speech_adaptation = speech.SpeechAdaptation(phrase_set_references=self.setting.phrase_sets)
                kwargs.update(
                    {"adaptation": speech_adaptation})

            # read audio file
            with io.open(audio_path, "rb") as audio:
                content = audio.read()
                audio = speech.RecognitionAudio(content=content)
                self.read_audio = True

            # transcribe audio file
            config = speech.RecognitionConfig(**kwargs)
            self.transcribing = True
            response = client.recognize(config=config, audio=audio)

        except Exception as e:
            logger.error(e, exc_info=e)
            self.transcribe_error = True
            raise TranscriptionError(str(e))
        else:
            logger.info("getting google response")
            self.transcribing = False
            self.transcribe_success = True
        return response

    def _prepare_utterance(
        self, response: cloud_speech.RecognizeResponse, offset=0.0
    ) -> List[Dict[str, str]]:
        """
        output the response data from google STT, convert the raw data to
        utterance data which is a list of dictionary in the format
        {speaker: , start_time: , end: , text: }

        Args:
            response
            offset

        Return:
            A list of dictionary that contains the utterance data of the
            audio file, each part of the audio file is stored in the format
            {speaker: , start_time: , end: , text: }
        """

        results = response.results
        """ Prepare Utterance """
        utterances = list()
        for result in results:
            alternative = result.alternatives[0]
            if alternative.transcript:
                for word in alternative.words:
                    utt = {
                        "start": word.start_time.total_seconds() + float(offset),
                        "end": word.end_time.total_seconds() + float(offset),
                        "text": word.word,
                        "speaker": str(word.speaker_tag),
                    }
                    utterances.append(utt)
        return utterances

    def _init_status(self):
        """
        Initializes the status
        """
        self.connected = False
        self.read_audio = False
        self.transcribing = False
        self.transcribe_success = False
        self.transcribe_error = False
        self.output_success = False

    def _transcribe_list_file(
        self, audios: List[str], workspace
    ) -> List[Dict[str, str]]:
        """transcribe a list of audios and return the utterance result

        Args:
            audios (List[str]): a list of path to audio source
            workspace (str): the path to workspace

        Returns:
            List[Dict[str, str]]: a list that represent the utterance result
        """
        utterances = []
        offset = 0.0
        for audio in audios:
            logger.info(f"transcribe {audio} in progress")
            response = self._run_engine(audio, workspace)
            assert response
            new_utt = self._prepare_utterance(response, offset)
            utterances.extend(new_utt)
            offset += self.current_chunk_duration
        return utterances

    @staticmethod
    def _get_chunk_duration(file_path: str, file_duration: int) -> int:
        """given the file path and file duration calculate the expected
           chunk duration that is able to send the chunked file to googl cloud

        Args:
            file_path (str): a str to the file path
            file_duration (int): the lenth of the file

        Returns:
            int: the expected duration of each audio chunk
        """
        duration = GOOGLE_CONFIG.maximum_duration
        filesize = os.path.getsize(file_path)
        num_chunks = file_duration // GOOGLE_CONFIG.maximum_duration
        size_60 = filesize / num_chunks
        if size_60 >= GOOGLE_CONFIG.maximum_size:
            duration = duration * (GOOGLE_CONFIG.maximum_size / size_60)
        return duration



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-09 11:30:48
# @Last Modified by:   Joanne Fan
# @Last Modified time: 2024-05-21 01:00:14

""" TODO:
1. only function to create and delete model are tested 
   function to train the model are not tested 
"""
# Standard library imports
from typing import Any, BinaryIO, List, TextIO, Dict, Callable, Union

# Local imports

# Third party imports
from gailbot.configs import watson_config_loader
from gailbot.shared.exception import transcribeException as ERR
from ibm_watson import SpeechToTextV1, ApiException
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_watson.speech_to_text_v1 import CustomWord

from gailbot.shared.utils.media import MediaHandler

import logging

logger = logging.getLogger(__name__)

from .codes import WatsonReturnCodes

WATSON_CONFIG = watson_config_loader()


class WatsonLMInterface:
    def __init__(self, apikey: str, region: str):
        self.connected_to_service = False
        self.apikey = apikey
        self.region = region
        self.media_h = MediaHandler()

        # Parse confs
        self._regions = WATSON_CONFIG.regions_uris
        self._format_to_content_types = WATSON_CONFIG.format_to_content
        self._defaults = WATSON_CONFIG.defaults
        self.max_size_bytes = WATSON_CONFIG.max_file_size_bytes

        if not self._is_api_key_valid(apikey, self._regions[region]):
            raise Exception(f"Apikey {apikey} invalid")
        if not region in self._regions:
            raise Exception(f"Region {region} not in {list(self._regions.keys())}")

        try:
            # Create the stt service and run
            authenticator = IAMAuthenticator(self.apikey)
            self.stt = SpeechToTextV1(authenticator=authenticator)
            self.stt.set_service_url(self._regions[self.region])
        except:
            raise Exception("Connect to STT failed")
        else:
            self.connected_to_service = True

    def get_base_model(self, model_name: str) -> Dict:
        """
        Obtain information regarding the given base language model.

        Args:
            model_name (str): Name of the language model.

        Returns:
            (Union[Dict[str,Any],None]):
                Dictionary with keys: name, language, rate, url,
                supported_features, description if successful.
                None if unsuccessful.
        """
        resp = self._execute_watson_method(
            self.stt.get_model, [WatsonReturnCodes.OK], [model_name]
        )
        return resp

    def get_base_models(self) -> List[str]:
        """
        Obtain a list of the names of base language models that are supported.

        Returns:
            (List[str]): Names of the base language models.
        """
        resp = self._execute_watson_method(self.stt.list_models, [WatsonReturnCodes.OK])
        return [model["name"] for model in resp["models"]] if "models" in resp else []

    def get_custom_model(self, customization_id: str) -> Dict:
        """
        Obtain information regarding a custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.

        Returns:
            (Union[Dict[str,Any],None]):
                Mapping from the following keys to their values:
                customization_id, CREATED, updated,language, dialect,
                versions, owner, name, description, base_model_name, status,
                progress
                None if unsuccessful.
        """
        resp = self._execute_watson_method(
            self.stt.get_language_model, [WatsonReturnCodes.OK], [customization_id]
        )
        return resp

    def get_custom_models(self) -> Dict[str, str]:
        """
        Get custom language models

        Returns:
            (Dict[str,str]):
                 Mapping from custom model name to the customization id.
        """
        custom_models = dict()
        resp = self._execute_watson_method(
            self.stt.list_language_models, [WatsonReturnCodes.OK]
        )
        for model in resp["customizations"]:
            custom_models[model["name"]] = model["customization_id"]
        return custom_models
    
    def get_custom_models_info(self, language: str = None) -> List[Dict[str, Any]]:
        """
        Get custom language models' detailed information including
        customization_id, created, updated, language, dialect, versions,
        owner, name, description, base_model_name, status, progress

        Returns:
            (List[Dict[str, Any]]):
                 Mapping from custom model name to the model information.
        """
        if language is None:
            resp = self._execute_watson_method(
                self.stt.list_language_models, [WatsonReturnCodes.OK]
            )
        else:
            resp = self._execute_watson_method(
                self.stt.list_language_models, [WatsonReturnCodes.OK], [], {"language": language}
            )

        return resp["customizations"]

######

    def get_custom_models_with_base(self, language: str) -> List[Dict[str, Any]]:
        """"
        Get all the base models that use the specified language

         Returns:
            (List[Dict[str, Any]]):
                Mapping from custom model name to the model information.
        """
        all_models = self.get_custom_models_info()
        resp = []
        for model in all_models:
            if model["base_model_name"] == language:
                name = model["name"]
                id = model["customization_id"]
                status = model["status"]
                resp.append({"name" : name, "customization_id" : id, "status": status})
        return resp


######
    def create_custom_model(self, name: str, base_model_name: str, description: str) -> Dict[str, Any]:
        """
        Create a new custom language model.
        Does NOT create a new model if one with the same name already exists.

        Args:
            name (str): Name of the model.
            base_model_name (str):
                Name of the base model. Must be a supported base model.
            description (str): Description of the model.

        Returns:
            (bool): True if successful. False otherwise.
        """
        if name in self.get_custom_models().keys():
            raise ERR.WatsonModelNameTaken
        resp = self._execute_watson_method(
            self.stt.create_language_model,
            [WatsonReturnCodes.CREATED],
            [name, base_model_name],
            {"description": description},
        )
        return resp

    def delete_custom_model(self, customization_id: str) -> bool:
        """
        Delete an existing custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.

        Returns:
            (bool): True if successful. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.delete_language_model, [WatsonReturnCodes.OK], [customization_id]
        )
        return resp == {}

    def train_custom_model(self, customization_id: str) -> bool:
        """
        Train a custom language model with a previously added resource.

        Args:
            customization_id (str): Unique ID of the custom language model.

        Returns:
            (bool): True if successful. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.train_language_model, [WatsonReturnCodes.OK], [customization_id]
        )
        return resp == {}

    def reset_custom_model(self, customization_id: str) -> bool:
        """
        Reset a custom language model to remove all loaded resources.

        Args:
            customization_id (str): Unique ID of the custom language model.

        Returns:
            (bool): True if successful. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.reset_language_model, [WatsonReturnCodes.OK], [customization_id]
        )
        return resp == {}

    def upgrade_custom_model(self, customization_id: str) -> bool:
        """
        Upgrade the base model of the custom language model to its latest
        version.

        Returns:
            (bool): True if successful. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.upgrade_language_model, [WatsonReturnCodes.OK], [customization_id]
        )
        return resp == {}

    def get_corpora(self, customization_id: str) -> List[Dict[str, Any]]:
        """
        Obtain the corpora used to train this custom model.

        Args:
            customization_id (str): Unique ID of the custom language model.

        Returns:
            (Union[List[Dict],None]):
                Each internal dict contains the keys: name,
                out_of_vocabulary_words,total_words, status.
                None if unsuccessful.
        """
        resp = self._execute_watson_method(
            self.stt.list_corpora, [WatsonReturnCodes.OK], [customization_id]
        )
        return resp["corpora"]

    def add_corpus(
        self, customization_id: str, corpus_name: str, corpus_file: BinaryIO
    ) -> bool:
        """
        Add a corpus to the specified custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            corpus_name (str): Name of the corpus
            corpus_file (BinaryIO): utf-8 encoded plain text file.
        """
        resp = self._execute_watson_method(
            self.stt.add_corpus,
            [WatsonReturnCodes.CREATED],
            [customization_id, corpus_name, corpus_file],
            {"allow_overwrite": True}
        )
        return resp == {}

    def delete_corpus(self, customization_id: str, corpus_name: str) -> bool:
        """
        Delete a corpus from a custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            corpus_name (str): Name of the corpus to delete.

        Returns:
            (bool): True if successful. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.delete_corpus,
            [WatsonReturnCodes.OK],
            [customization_id, corpus_name],
        )
        return resp == {}

    def get_corpus(self, customization_id: str, corpus_name: str) -> Dict[str, Any]:
        """
        Obtain information about a specific custom corpus used to train the
        custom model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            corpus_name (str): Name of the corpus.

        Returns:
            (Union[Dict[str,Any],None]): Contains the keys: "name",
            "out_of_vocabulary_words", "total_words","status"
            None if unsuccessful.
        """
        resp = self._execute_watson_method(
            self.stt.get_corpus, [WatsonReturnCodes.OK], [customization_id, corpus_name]
        )
        return resp

# TODO: methods on words are not used currently so the below methods may be outdated.
    def get_custom_words(self, customization_id: str) -> List[Dict[str, Any]]:
        """
        Obtain all the custom words used to train a custom model.

        Args:
            customization_id (str): Unique ID of the custom language model.

        Returns:
            (Union[List[Dict], None]):
                Each dictionary contains the keys: "word","sounds_like",
                "display_as","count","source"
                None if unsuccessful.
        """
        resp = self._execute_watson_method(
            self.stt.list_words, [WatsonReturnCodes.OK], [customization_id]
        )
        return resp["words"]

    def add_custom_words(self, customization_id: str, words: List[str]) -> bool:
        """
        Add one or more custom words to the specified custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            words (List[str]): List of words to add to the model.

        Returns:
            (bool): True if successful. False otherwise.
        """
        custom_words = list()
        for word in words:
            custom_words.append(CustomWord(word=word))
        resp = self._execute_watson_method(
            self.stt.add_words,
            [WatsonReturnCodes.CREATED],
            [customization_id, custom_words],
        )
        return resp

    def delete_custom_words(self, customization_id: str, word: str) -> bool:
        """
        Delete a word from a custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            word (str): Word to delete.

        Returns:
            (bool): True if successful. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.delete_word, [WatsonReturnCodes.OK], [customization_id, word]
        )
        return resp

    def get_custom_grammars(self, customization_id: str) -> List[Dict]:
        """
        Obtain a list of the grammars of a custom model.

        Args:
            customization_id (str): Unique ID of the custom language model.

        Returns:
            (Union[List[Dict], None]):
                Internal dictionaries contain the keys:
                out_of_vocabulary_words","name","status"
                None of unsuccessful.
        """
        resp = self._execute_watson_method(
            self.stt.list_grammars, [customization_id]
        )
        return resp["grammars"]

    def get_custom_grammar(self, customization_id: str, grammar_name: str) -> Dict:
        """
        Obtain information about a specific grammar in a custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            grammar_name (str): Name of the custom grammar.

        Returns:
            (Union[Dict, None]):
                Contains the keys: "out_of_vocabulary_words","name","status"
                None of unsuccessful.
        """
        resp = self._execute_watson_method(
            self.stt.get_grammar,
            [WatsonReturnCodes.OK],
            [customization_id, grammar_name],
        )
        return resp

    def add_custom_grammar(
        self,
        customization_id: str,
        grammar_name: str,
        grammar_file: TextIO,
        content_type: str,
    ) -> bool:
        """
        Add a grammar to the custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            grammar_name (str): Name of the custom grammar.
            grammar_file (TextIO): Plain text file containing grammar.
            content_type (str):
                Format of the grammar. one of:
                    application/srgs, application/srgs+xml

        Returns:
            (bool): True if successfully added. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.add_grammar,
            [WatsonReturnCodes.CREATED],
            [customization_id, grammar_name, grammar_file, content_type, True],
        )
        return resp

    def delete_custom_grammar(self, customization_id: str, grammar_name: str) -> bool:
        """
        Delete the specific grammar from a custom language model.

        Args:
            customization_id (str): Unique ID of the custom language model.
            grammar_name (str): Name of the custom grammar.

        Returns:
            (bool): True if successfully added. False otherwise.
        """
        resp = self._execute_watson_method(
            self.stt.delete_grammar,
            [WatsonReturnCodes.OK],
            [customization_id, grammar_name],
        )
        return resp

    # Others
    def _execute_watson_method(
        self,
        method: Callable,
        expected_response_codes: List = [],
        args: List = [],
        kwargs: Dict = {},
    ) -> Dict:
        """
        Execute a watson method only if connected to watson.

        Args:
            method (Callable): Method to execute.
            expected_response_codes (List[WatsonReturnCodes]):
                Watson codes that are expected for a successful response.
            args (List): Arguments to pass to the method.
            kwargs (Dict): Keyword arguments to method

        Returns:
            (Dict):
                response from watson if successful.
                Raises exception otherwise.
        """
        if not self.connected_to_service:
            raise ERR.InternetConnectionError
        try:
            resp = method(*args, **kwargs)
            status_code = resp.get_status_code()
            if any(
                [
                    status_code == code
                    for code in expected_response_codes
                ]
            ):
                return resp.get_result()
            raise Exception(f"Received wrong status code {status_code}.")
        except Exception as e:
            logger.error(f"Exception raised: {e}", exc_info=e)
            raise ERR.WatsonMethodExecutionError

    def _is_api_key_valid(self, apikey: str, url: str) -> bool:
        """
        Determine if the given apikey is valid.

        Args:
            apikey (str): API key for the watson STT service.

        Returns:
            (bool): True if the key is valid. False otherwise.
        """
        try:
            stt = self._initialize_stt_service(apikey)
            stt.set_service_url(url)
            stt.list_models()
            return True
        except:
            return False

    def _initialize_stt_service(self, apikey: str) -> SpeechToTextV1:
        """
        Initialize a SpeechToTextV1 object given an apikey.

        Args:
            apikey (str): Valid API key for the watson STT service.

        Returns:
            (SpeechToTextV1)
        """
        authenticator = IAMAuthenticator(apikey)
        stt = SpeechToTextV1(authenticator=authenticator)
        return stt



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-07-24 14:34:21
from dataclasses import dataclass
from dict_to_dataclass import field_from_dict, DataclassFromDict
from typing import Dict
import toml


@dataclass
class WatsonConfig(DataclassFromDict):
    """Loads data from the Watson engine configuration"""

    max_file_size_bytes: float = field_from_dict()
    regions_uris: Dict = field_from_dict()
    format_to_content: Dict = field_from_dict()
    defaults: Dict = field_from_dict()
    name: str = field_from_dict()
    workspace: str = field_from_dict()


def load_watson_config(path: str):
    """Loads data from the Watson engine configuration

    Args:
        path (str) : path to the toml file to load
    """
    d = toml.load(path)
    return WatsonConfig.from_dict(d["watson"])



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-08 12:54:35
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-02-06 16:10:12
# @Description: Component is a wrapper around an executable function which
# will be accepted by Pipeline


from abc import ABC
from enum import Enum
from dataclasses import dataclass
from typing import Any, Dict


class ComponentState(Enum):
    """
    Class containing the status of a component object.
    """

    READY = 0
    SUCCESS = 1
    FAILED = 2


@dataclass
class ComponentResult:
    """
    Class containing the result of a component object.
    """

    state: ComponentState = ComponentState.FAILED
    result: Any = None
    runtime: float = 0


class Component(ABC):
    """
    Wrapper for a function that is run in the transcriptionPipeline.
    Should be subclassed.
    """

    def __init__(self, *args, **kwargs):
        pass

    def __repr__(self):
        raise NotImplementedError()

    def __call__(
        self, dependency_outputs: Dict[str, ComponentResult], *args, **kwargs
    ) -> ComponentState:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# @Author: Muhammad Umair
# @Date:   2023-01-08 15:53:06
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-11 00:18:26


from dataclasses import dataclass
from concurrent.futures import ThreadPoolExecutor, Future, wait
from typing import Tuple, Callable, List, Dict, Any, Union
from gailbot.shared.utils.logger import makelogger

logger = makelogger("thread")


#### Custom Exceptions for Threads ####
class TaskNotFoundException(Exception):
    def __init__(self, task_key=None) -> None:
        self.task_key = task_key

    def __str__(self) -> str:
        return "task" + str(self.task_key) + " is not found"


class TaskNotFinishedException(Exception):
    def __init__(self, task_key=None) -> None:
        self.task_key = task_key

    def __str__(self) -> str:
        return "task" + str(self.task_key) + " is not finished"


class TaskCancelException(Exception):
    def __init__(self, task_key=None) -> None:
        self.task_key = task_key

    def __str__(self) -> str:
        return "task" + str(self.task_key) + " cannot be canceled"


class TaskCreateError(Exception):
    def __init__(self, msg=None) -> None:
        self.msg = msg

    def __str__(self) -> str:
        return "task cannot be created due to the error " + self.msg


class ThreadError(Exception):
    def __init__(self, task_key=None, msg: str = None) -> None:
        self.task_key = task_key
        self.msg = msg

    def __str__(self) -> str:
        return "task" + str(self.task_key) + " has an error " + self.msg


### Defines different possible statuses for a task ###
@dataclass
class Status:
    running = "RUNNING"
    pending = "PENDING"
    finished = "FINISHED"
    cancelled = "CANCELLED"
    error = "ERROR"


class ThreadPool(ThreadPoolExecutor):
    """
    Implement a threadpool that is able to run tasks parallel on different
    threads, support functions to query tasks status, return task result and
    add callback task to previous task

    Inheritance:
    ThreadPoolExecutor

    """

    def __init__(self, max_workers: int, *args, **kwargs) -> None:
        """
        Constructs a threadpool with the given size that is able to run tasks on different
            threads.

        Args:
            max_workers (int):  the maximum number of tasks that will be run
                                at the same time parallel
        """
        super().__init__(max_workers, *args, **kwargs)
        self.num_thread = max_workers
        self.task_pool: Dict[Any, Future] = (
            dict()
        )  # used to keeps track of the task status
        self.next_key = 0  # next available key

    def get_num_threads(self) -> int:
        """
        Returns the number of threads available in the thread pool.

        Return:
            int: Integer value representing the number of
                 threads in the given thread pool.
        """
        return self.num_thread

    def add_task(
        self,
        task: Callable,
        args: List = [],
        kwargs: Dict = {},
        error_fun: Callable = None,
        key: Union[str, int] = None,
    ) -> Union[str, int]:
        """
        Adds the given task to the task pool.

        Args:
            fun: Callable: function that will be run in the threadpool
            args: List (Optional): a list argument to the function
            kwargs: Dict (Optional): a list of key word arguments to the function
            error_fun: Callable (Optional): a function to handle the error
            key: Union[str, int](Optional): a user provided key that can be used to query
                                the task progress and result after submitting it
                                to the threadpool, caller is responsible for
                                providing unique key
        Return:
            Union[str, int]: key that will be used to keep track of the task in the thread.
        """
        try:
            if not callable(task) or type(args) != list or type(kwargs) != dict:
                logger.error(f"type error in adding thread task ")
                raise TypeError("Type error in adding thread task.")
            worker = self.submit(task, *args, **kwargs)
            if key:
                self.task_pool[key] = worker
                return key
            else:
                self.task_pool[self.next_key] = worker
                self.next_key += 1
                return self.next_key - 1
        except Exception as e:
            logger.error(e, exc_info=e)
            if error_fun:
                error_fun()
            else:
                raise TaskCreateError(e)

    def check_task_status(self, key: Union[str, int]):
        """
        Return a string representing the status of the current task.

        Args:
            key: key to find the given task in the thread.

        Returns:
            String representing the status of the current task.
        """
        self._task_in_pool(key)
        future: Future = self.task_pool[key]
        return future._state

    def get_tasks_with_status(self, status: Status) -> List[Tuple[int, str]]:
        """
        Returns a list of tasks that match the status passed in by caller

        Args:
            status (Status): Status ot the task

        Returns:
            A List[Tuple[int, str]] of tasks in the thread pool that
                match the requested status.
        """
        return [
            task_key
            for task_key in self.task_pool.keys()
            if self.check_task_status(task_key) == status
        ]

    def get_task_result(self, key: Union[str, int], error_fun: Callable = None):
        """
        Gets the result from the given task.

        Args:
            key (int): Key to find the given task in the thread pool.

        Return:
            If result is successfully obtained, return the result of the given task.
            else return false as default or call the error handling function passed
            in from user
        """
        self._task_in_pool(key)
        try:
            future = self.task_pool[key]
            return future.result()
        except Exception as e:
            logger.error(
                f"task with task key {key} with future object {future} received an exception {e} , the function exception is {future.exception()}",
                exc_info=e,
            )
            if error_fun and callable(error_fun):
                logger.error(e)
                error_fun()
                return future.result()
            else:
                raise ThreadError(key, f"received exception {e}")  ##

    def completed(self, key: Union[str, int], error_fun: Callable = None) -> bool:
        """
        Checks if a given task has been completed.

        Args:
            key (int): Key to find the given task in the thread pool.
            error_fun: Callable function used if there is an error when
                       executing the task.

        Raise:
            Raise ThreadError if the result of the given task is not
            successfully obtained.

        Return:
            bool: True if the given task has been completed, false if not.
        """
        self._task_in_pool(key)
        try:
            return self.task_pool[key].done()
        except:
            if error_fun and callable(error_fun):
                error_fun()
            else:
                raise ThreadError(key, "ERROR: failed to get thread status")

    def wait_for_all_completion(self, error_fun: Callable = None):
        """
        Wait for all tasks in the thread pool to be completed.

        Args:
            error_fun: Callable function used if there is an error when
                       executing the task.

        Raises:
            Raises a ThreadError if the function cannot be properly executed.

        Return:
            Calls wait_for_task() for all tasks in the thread pool.
        """
        futures = self.task_pool.values()
        wait(futures)

    def wait_for_task(self, key: Union[str, int], error_fun: Callable = None):
        """
        Waits for a certain task in the thread pool to be completed.

        Args:
            key (int): Key to find the given task in the thread pool.
            error_fun: Callable function used if there is an error when
                       executing the task.

        Raises:
            Raises a ThreadError if the task cannot be properly executed.

        Return:
            None
        """
        self._task_in_pool(key)
        future = self.task_pool[key]
        try:
            wait([future])
            assert not future.exception()
        except Exception as e:
            logger.error(e, exc_info=e)
            if error_fun:
                error_fun()
            else:
                raise ThreadError(key, f"received exception {e}")

    def cancel(self, key: Union[str, int]) -> bool:
        """
        Cancels the task at the given key.

        Args:
            key (int): Key to find the given task in the thread pool.

        Raises:
            TaskCancelException if the task at the given key cannot be
            properly cancelled.

        Return:
            true if the thread is canceled
        """
        self._task_in_pool(key)
        try:
            self.task_pool[key].cancel()
            if not self.task_pool[key].cancelled():
                raise TaskCancelException(key)
            return True
        except:
            return False

    def cancel_all(self) -> None:
        """
        Cancels all the running tasks in the thread pool.

        Raises:
            TaskCancelException if the current task cannot be properly cancelled.

        Return:
            None
        """
        try:
            for task in self.task_pool.values():
                if not task.done():
                    task.cancel()
            return True
        except:
            return False

    def add_callback_with_arg(
        self, key: Union[str, int], fun: Callable, error_fun: Callable = None
    ):
        """
        Adds a function to the thread as a callback of a previous function

        Args:
            key (int): key to find the given task.
            fun (Callable): function to be called as the callback.
            error_fun: Callable function used if there is an error
                       when executing the task.

        Raises:
            ThreadError if the callback function cannot be added

        Return:
            Returns the key that identify the newly added task
        """
        self._task_in_pool(key)
        future = self.task_pool[key]
        res = [future.result()]
        if res and future.done():
            return self.add_task(fun, res)
        elif error_fun and callable(error_fun):
            error_fun()
        else:
            raise ThreadError(key, "ERROR: Failed to add callback")

    def add_callback(self, key: Union[str, int], fun: Callable):
        """add call back function fun that will run after the task identified
            by key is finished running

        Args:
            key (Union[str, int]): the key that identifies the existing an task
            fun (Callable): the call back function that will be added
        """
        if not callable(fun):
            raise ThreadError(key, "ERROR: Non-function added as callback")
        future = self.task_pool[key]
        future.add_done_callback(fun)

    def add_callback_on_result(
        self,
        key: Union[str, int],
        fun: Callable,
        args: List = [],
        kwargs: Dict = {},
        error_fun: Callable = None,
    ) -> int:
        """
        Adds a task to the thread after one task has finished running.

        Args:
            key (int): Key to find the current task.
            fun (Callable): function to be called after the current task has
                            been completed.
            error_fun: Callable function used if there is an error when
                      executing the task.

        Raises:
            Raises ThreadError if the task identified by key is throws an
            exception

            Raises TypeError if the arguments is not passed in correctly

        Returns:
            Returns the key that identify the newly added task
        """
        self._task_in_pool(key)
        future = self.task_pool[key]
        wait([future])
        if not callable(fun) and type(args) == list and type(kwargs) == dict:
            logger.error(f"type error in adding thread task ")
            raise TypeError(
                f"Type error in adding callback on result to thread task {key}"
            )
        try:
            future.result()
            return self.add_task(fun, args, kwargs)
        except Exception as e:
            logger.error(e, exc_info=e)
            if error_fun:
                error_fun()
            else:
                raise ThreadError(
                    key, f"ERROR: Failed to add callback task, get error {e}"
                )

    def is_busy(self) -> bool:
        """
        Determines if the thread queue has tasks in it.

        Returns:
            bool: True if the queue has non-zero tasks, false if otherwise.
        """
        return self._work_queue.qsize() != 0

    def count_waiting_tasks(self) -> int:
        """
        Calculates the total number of tasks waiting in the queue.

        Returns:
            Integer representing the total waiting tasks
        """
        return self._work_queue.qsize()

    def count_completed_tasks(self) -> int:
        """get the number of completed task

        Returns:
            int: the number of completed task
        """
        num = 0
        for future in self.task_pool.values():
            if future.done():
                num += 1
        return num

    def count_total_tasks(self) -> int:
        """get the number of all the tasks that has been added to the threadpool,
           which includes failure task, completed task, running task and pending task


        Returns:
            int:  the total number of task
        """
        return len(self.task_pool)

    ############ private function  ##########
    def _task_in_pool(self, key: int) -> None:
        """
        Private function to determine if the given task is currently in the task pool.

        Args:
            key (int): key with which to find the given task in the thread.

        Raises:
            Raises TaskNotFoundException if the given task is not currently
            in the thread.

        Return:
            None
        """
        if not key in self.task_pool:
            raise TaskNotFoundException(key)



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-01-29 13:26:17
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-04-10 22:00:14
# @Description: Define the schema for data related to a single plugin suite
from typing import Dict, List, Optional
from pydantic import BaseModel


class Suite(BaseModel):
    name: str
    id: str
    description: str
    Version: str
    Author: str
    plugins: str


    # Version: str
    # Author: str
    # Email: str
    # BucketName: Optional[str] = ""
    # ObjectName: Optional[str] = ""
    # ObjectUrl: Optional[str] = ""


class Requirements(BaseModel):
    requirements: Optional[Dict[str, str]] = {}  

class Dependencies(BaseModel):
    dependencies: Optional[Dict[str, str]] = {}  

class ConfModel(BaseModel):
    # """dictionary type for plugin suite configuration dictionary"""

    suite: Suite
    requirements: Optional[Requirements] = None  
    dependencies: Optional[Dependencies] = None 

    # suite_name: str
    # plugins: List[PluginDict]



# -*- coding: utf-8 -*-
# @Author: Vivian Li
# @Date:   2024-02-24 10:09:53
# @Email:
# @Last Modified by:   Your name
# @Last Modified time: 2024-03-06 18:53:42
# @Descriptions


from .payloadObject import PayloadObject
from .untranscribedDirPayload import untranscribedDirPayload
from .transcribedDirPayload import TranscribedDirPayload



# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-07-17 19:21:51
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-08-01 05:34:30
from typing import List
from abc import ABC
from gailbot.pluginSuiteManager.suite.suite import PluginSuite


class PluginSuiteLoader(ABC):
    """base class for plugin loader"""

    def load(self, *args, **kwargs) -> List[PluginSuite]:
        """
        Returns:
            a list of valid plugin suite, return empty list if no valid plugin suite is loaded
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# @Author: Hannah Shader, Jason Wu, Jacob Boyar
# @Date:   2023-06-27 12:16:07
# @Last Modified by:   Vivian Li
# @Last Modified time: 2024-02-17 08:57:01
from gailbot import Plugin
from gailbot import GBPluginMethods
from typing import Dict, Any
from .testOne import TestOne

###############################################################################
# CLASS DEFINITIONS                                                           #
###############################################################################
class TestThree(Plugin):
    """
    Wrapper class for the Gaps plugin. Contains functionality that inserts
    gap markers
    """

    def __init__(self) -> None:
        super().__init__()
        """
        Initializes the gap plugin

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

    def apply(self, dependency_outputs: Dict[str, Any], methods: GBPluginMethods):
        """
        Parameters
        ----------
        dependency_outputs: a list of dependency outputs
        methods: the methods being used, currently GBPluginMethods

        Returns
        -------
        A structure interact instance
        """
        self.successful = True
        return "TestThree"
